from http import HTTPStatus
from typing import Annotated, override

from fastapi import Response, Security
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from phederation.api.routes.base import BaseRoute
from phederation.federation.discovery import InstanceInfo
from phederation.models import (
    APActivity,
    ErrorMessageActivityHandler,
    ErrorMessageActorUnauthorized,
    RelativeLink,
)
from phederation.security.authentication import UserInfo
from phederation.utils import NodeInfo, ObjectId
from phederation.utils.base import UrlType, assemble_id_url
from phederation.utils.exceptions import ActivityPubException, ResolverError


class InstanceRoute(BaseRoute):

    @override
    def setup(self):
        @self.router.get("/health", tags=["instance"])
        def health_check():  # pyright: ignore[reportUnusedFunction]
            return {"status": "ok", "description": "The API is available"}

        @self.router.get("/robots.txt", tags=["instance"])
        def get_robots_txt():  # pyright: ignore[reportUnusedFunction]
            """Get robots allow/disallow file."""
            with open("settings/robots.txt", "r") as robots_txt_file:
                return Response(
                    content="\n".join(robots_txt_file.readlines()),
                    status_code=HTTPStatus.OK,
                    media_type="text/plain",
                )

        @self.router.get("/.well-known/instanceinfo", tags=["instance"])
        async def get_instanceinfo() -> InstanceInfo:  # pyright: ignore[reportUnusedFunction]
            """Get NodeInfo links."""
            return await self.server.instance.get_instance_info()

        @self.router.get("/.well-known/nodeinfo", tags=["instance"])
        def get_nodeinfo_links() -> dict[str, list[RelativeLink]]:  # pyright: ignore[reportUnusedFunction]
            """Get NodeInfo links."""
            return self.server.instance.get_nodeinfo_links()

        @self.router.get("/nodeinfo/2.1", tags=["instance"])
        async def get_nodeinfo() -> NodeInfo:  # pyright: ignore[reportUnusedFunction]
            """Get NodeInfo."""
            return await self.server.instance.get_nodeinfo()

        @self.router.get("/nodeinfo/{primary}", tags=["instance"])
        async def get_nodeinfo_from_primary(primary: str):  # pyright: ignore[reportUnusedFunction]
            """Get NodeInfo stored in storage."""
            nodeinfo_id = assemble_id_url(type=UrlType.NodeInfo, base_url=self.api.settings.domain.base_url, primary=primary)
            node_info = await self.server.storage.node_info.read(id=nodeinfo_id)
            if not node_info:
                return JSONResponse(
                    content={"status": "error", "description": "Node info document not found"},
                    status_code=HTTPStatus.NOT_FOUND,
                )
            node_info.id = nodeinfo_id
            return node_info

        @self.router.get("/ns/schema/2.1", tags=["instance"])
        def get_nodeinfo_schema():  # pyright: ignore[reportUnusedFunction]
            """Get NodeInfo schema."""
            return JSONResponse(
                content=jsonable_encoder(self.server.instance.get_nodeinfo_schema()),
                status_code=HTTPStatus.OK,
            )

        @self.router.post(
            "/inbox",
            tags=["instance", "users"],
            responses={
                HTTPStatus.INTERNAL_SERVER_ERROR: {
                    "model": ErrorMessageActivityHandler,
                    "description": ErrorMessageActivityHandler().description,
                },
                HTTPStatus.UNAUTHORIZED: {
                    "model": ErrorMessageActorUnauthorized,
                    "description": ErrorMessageActorUnauthorized().description,
                },
            },
        )
        async def post_shared_inbox(activity: APActivity):  # pyright: ignore[reportUnusedFunction]
            """Handle shared inbox."""
            # Handle activity
            try:
                self.logger.debug(f"Shared inbox received activity. Processing.")

                # handle the shared inbox
                result = await self.server.handle_inbox(activity=activity)
            except ActivityPubException as e:
                self.logger.debug(f"Shared inbox post fails. Error: '{e.message}'. Failing silently.")
                return JSONResponse(
                    content={"Status": "Error", "Description": (e.user_message or "Unknown error") + "; Failing silently"},
                    status_code=HTTPStatus.CONTINUE,
                    media_type="application/activity+json",
                )

            result_id: ObjectId | None = result.get_success_id()
            if result.success and result_id:
                return JSONResponse(
                    content={"Status": "OK"},
                    status_code=HTTPStatus.CREATED,
                    media_type="application/activity+json",
                    headers={"Location": result_id},
                )
            else:
                self.logger.warning(f"Instance shared inbox processing error: {result}")
                return JSONResponse(
                    status_code=HTTPStatus.BAD_REQUEST,
                    content={"Status": "error", "Description": "Shared inbox processing failed"},
                    media_type="application/activity+json",
                )

        @self.router.post(
            f"/{UrlType.Resolve.value}",
            tags=["instance"],
            responses={
                HTTPStatus.INTERNAL_SERVER_ERROR: {
                    "model": ErrorMessageActivityHandler,
                    "description": ErrorMessageActivityHandler().description,
                },
                HTTPStatus.UNAUTHORIZED: {
                    "model": ErrorMessageActorUnauthorized,
                    "description": ErrorMessageActorUnauthorized().description,
                },
            },
        )
        async def post_resolve_ids(  # pyright: ignore[reportUnusedFunction]
            ids_to_resolve: list[ObjectId],
            user_info: Annotated[UserInfo | None, Security(self.middleware.authentication, scopes=["user"])],
        ):
            """Resolve multiple objects from this instance in a json list."""
            try:
                self.logger.debug(f"Resolve endpoint received {len(ids_to_resolve)} ids. Processing.")
                if "resolve" not in self.server.settings.federation.endpoints:
                    raise ResolverError("Endpoint 'resolve' is not activated on this instance, cannot resolve multiple ids at once")
                if len(ids_to_resolve) > self.server.settings.federation.max_concurrent_resolve:
                    raise ResolverError(
                        f"Endpoint 'resolve' can only resolve {self.server.settings.federation.max_concurrent_resolve} ids at once (requested: {len(ids_to_resolve)})"
                    )

                result = await self.server.resolve_multiple(ids_to_resolve, actor_for_authorization=user_info.actor_id if user_info else None)
                self.logger.debug(f"Resolved {len(result)} objects out of {len(ids_to_resolve)}")
            except ActivityPubException as e:
                self.logger.debug(f"Resolve endpoint POST failed. Error: '{e.message}'. Failing silently.")
                return JSONResponse(
                    content={"Status": "Error", "Description": (e.user_message or "Unknown error") + "; Failing silently"},
                    status_code=HTTPStatus.CONTINUE,
                    media_type="application/activity+json",
                )

            return JSONResponse(
                content=[obj.serialize() for obj in result],
                status_code=HTTPStatus.CREATED,
                media_type="application/activity+json",
            )



"""
Federation module for ActivityPub server-to-server interactions.
"""

from .client import ActivityPubClient
from .delivery import ActivityDelivery, DeliveryResult
from .discovery import InstanceDiscovery
from .instance import APInstance, InstanceCollectionType
from .middleware import ActivityPubMiddleware
from .resolver import ActivityPubResolver
from .server import ActivityPubServer

__all__ = [
    "APInstance",
    "InstanceCollectionType",
    "ActivityPubClient",
    "ActivityDelivery",
    "DeliveryResult",
    "ActivityPubResolver",
    "ActivityPubServer",
    "InstanceDiscovery",
    "ActivityPubMiddleware",
]



import asyncio
from collections.abc import Awaitable
from dataclasses import dataclass, field
from datetime import datetime, timezone
from http import HTTPStatus
import json
from json import JSONDecodeError
from logging import Logger
from typing import Any, BinaryIO, Callable, cast, override
from urllib.parse import urlparse

from fastapi.datastructures import UploadFile
import httpx
from pydantic import SecretStr
from urllib3.util import Url

from phederation.federation.discovery import InstanceDiscovery
from phederation.federation.instance import APInstance
from phederation.federation.ratelimit import RateLimiter
from phederation.federation.resolver import ActivityPubResolver
from phederation.models import APObject, CRUD, ValidCollection, dereference
from phederation.models.activities import APActivity, ActivityInQueue
from phederation.models.proofs import to_sha256
from phederation.security.http_signatures import HTTPSignatureVerifier
from phederation.security.key_management import KeyManager
from phederation.storage.base import StorageBackend
from phederation.utils import UrlType, validate_object_id
from phederation.utils.base import AccessType, ActivityType, ObjectId
from phederation.utils.exceptions import (
    ActivityPubException,
    AuthorizationError,
    DecodingError,
    DeliveryError,
    ResolverError,
    StorageError,
)
from phederation.utils.logging import configure_logger
from phederation.utils.serialization import to_json_string
from phederation.utils.settings import DeliveryQueueType, PhedSettings
from phederation.utils.version import PHEDERATION_HEADERS


@dataclass
class DeliveryResult:
    """Result of delivering a message to one or more recipients.
    Contains information for each recipient in success/failed lists."""

    success: list[str | dict[str, Any]] = field(default_factory=lambda: [])
    """Lists the responses for successful deliveries."""

    failed: list[str | dict[str, Any]] = field(default_factory=lambda: [])
    """Lists the responses for failed deliveries."""

    status_code: None | int = None
    """Combined status code. Is only set to a positive message if all recipients could be reached, otherwise to an error code."""

    error_message: None | str = None
    """Combined error message for all failed deliveries."""

    @staticmethod
    def OK():
        return DeliveryResult(success=["OK"], status_code=HTTPStatus.OK)

    def __post_init__(self):
        self.success = self.success or []
        self.failed = self.failed or []

    @override
    def __str__(self):
        return to_json_string({"success": self.success, "failed": self.failed, "status_code": self.status_code, "error_message": self.error_message})

    @override
    def __repr__(self):
        return str(self)

    def get_success_id(self) -> ObjectId | None:
        """Returns the ObjectId of a successful delivery.
        Used mostly when delivering a Create activity and requesting the id of the activity as it was created on the instance.

        Returns:
            ObjectId | None: Id if successful, None otherwise.
        """
        if self.success:
            first_success = self.success[0]
            if isinstance(first_success, dict) and "Location" in first_success:
                id: str | None = first_success.get("Location", None)
                if id:
                    return id
            if isinstance(first_success, str):
                try:
                    return first_success
                except Exception:
                    return None
        return None

    def get_success_id_or_raise(self) -> ObjectId:
        id = self.get_success_id()
        if not id:
            raise DeliveryError("No id found in DeliveryResult")
        return id

    @staticmethod
    def create_from_list(delivery_results: list["DeliveryResult"]) -> "DeliveryResult":
        """Combine a list of DeliveryResult objects into a single one.
        This will combine the error_message strings by joining them, and discard any status_message.

        Args:
            delivery_results (list[DeliveryResult]): list of results.

        Returns:
            DeliveryResult: The combined result object.
        """
        result = DeliveryResult()
        if len(delivery_results) == 0:
            result.error_message = "Nothing delivered."
            result.status_code = HTTPStatus.BAD_REQUEST
            return result
        result.success = []
        result.failed = []
        result.status_code = None
        for d in delivery_results:
            result.success += d.success
            result.failed += d.failed
        result.error_message = "; ".join([d.error_message for d in delivery_results if d.error_message])
        # status codes: if list only contains one item, set to that one. If at least one error, set to BAD_REQUEST
        if len(delivery_results) == 1:
            result.status_code = delivery_results[0].status_code
        elif result.error_message and len(result.error_message) > 0:
            result.status_code = HTTPStatus.BAD_REQUEST
        return result


class APDeliveryQueue:
    """The message queue of all attempted deliveries that have not been sent yet.

    Stores all activities in storage and then pulls them out sequentially, ordered by their to_sent date.
    """

    def __init__(self, storage: StorageBackend, maxsize: int = 0) -> None:
        self.maxsize: int = maxsize
        self.storage: StorageBackend = storage
        self.message_queue: CRUD[ActivityInQueue]

    async def initialize(self):
        self.message_queue = self.storage.table(model_class=ActivityInQueue, name=UrlType.MessageQueue.value)

    async def size(self):
        return await self.message_queue.count()

    async def empty(self):
        return await self.size() == 0

    async def full(self):
        return self.maxsize > 0 and await self.size() >= self.maxsize

    async def put(self, activity: APActivity, recipients: list[ObjectId], send_at: datetime | None = None):
        if not send_at:
            send_at = datetime.now(timezone.utc)
        activity_to_send = ActivityInQueue(id=activity.id, recipients=recipients, activity=activity, send_at=send_at)
        return await self.message_queue.create(activity_to_send, raise_if_exists=False)

    @staticmethod
    def _sort_by_date(activity_to_send: ActivityInQueue):
        return activity_to_send.send_at

    async def pop(self):
        queue = await self.message_queue.select()
        if len(queue) > 0:
            queue.sort(key=APDeliveryQueue._sort_by_date)
            first_activity = queue[0]
            if not first_activity.id:
                raise StorageError("Activity in queue does not have an id assigned")
            _ = await self.message_queue.delete(id=first_activity.id)
            return first_activity
        else:
            return None

    async def clear(self):
        items = await self.message_queue.select()
        _ = await self.message_queue.delete([item.id for item in items if item.id is not None])


class ActivityDelivery:
    """Activity delivery across the federated network.

    Features:
     - Deliver activities immediately or with temporalio workflow.
     - Cleanup activities before sending them.
     - Detect local delivery and switch to faster delivery, bypassing the network.
    """

    def __init__(
        self,
        discovery: InstanceDiscovery,
        resolver: ActivityPubResolver,
        settings: PhedSettings,
        rate_limiter: RateLimiter,
        instance: APInstance | None = None,
        key_manager: KeyManager | None = None,
    ):
        """Create a new ActivityDelivery.

        Args:
            discovery: To obtain URLs for instances and actors.
            resolver: To resolve actor ids.
            settings: To configure timeouts, retries, delays etc.
            delivery_workflow: For concurrent delivery in temporalio.
            rate_limiter: Rate limiting the delivery to not overload remote instances.
            instance: To access information on blocked and federated instances.
            key_manager: For authentication.
        """
        self.key_manager: KeyManager | None = key_manager
        self.discovery: InstanceDiscovery = discovery
        self.resolver: ActivityPubResolver = resolver
        self.instance: APInstance | None = instance
        self.settings: PhedSettings = settings
        self.local_delivery_inbox: Callable[[APActivity, ObjectId | None], Awaitable[DeliveryResult]] | None = None
        self.local_delivery_outbox: Callable[[APActivity | APObject], Awaitable[DeliveryResult]] | None = None
        self.delivery_workflow: Callable[[], Awaitable[DeliveryResult]] | None = None
        self.delivery_queue: APDeliveryQueue | None = APDeliveryQueue(storage=self.key_manager.storage) if self.key_manager else None
        self.signature_verifier: HTTPSignatureVerifier | None = HTTPSignatureVerifier(key_manager=key_manager) if key_manager else None
        self.rate_limiter: RateLimiter = rate_limiter
        # this can be set by the test suite to replace sending actual http requests and instead use the test_client
        self.requests: httpx.AsyncClient = httpx.AsyncClient()
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

    async def initialize(
        self,
        local_delivery_inbox: Callable[[APActivity, ObjectId | None], Awaitable[DeliveryResult]],
        local_delivery_outbox: Callable[[APActivity | APObject], Awaitable[DeliveryResult]],
        delivery_workflow: Callable[[], Awaitable[DeliveryResult]] | None = None,
    ) -> None:
        """Initializes the delivery with local routing for inbox and outbox (from `ActivityPubServer`).
        Also enables to pass a temporalio delivery workflow method to send activities to a list of recipients.
        """
        self.local_delivery_inbox = local_delivery_inbox
        self.local_delivery_outbox = local_delivery_outbox
        self.delivery_workflow = delivery_workflow
        if self.delivery_queue:
            await self.delivery_queue.initialize()

    async def check_is_local(self, actor_id: str) -> bool:
        parsed_url = urlparse(actor_id)
        actor_domain = validate_object_id(Url(scheme=parsed_url.scheme, host=parsed_url.netloc).url)
        return actor_domain.lower() == self.settings.domain.base_url.lower()

    @staticmethod
    async def remove_hidden_recipients(activity: APActivity) -> APActivity:
        """Make sure that both bto and bcc fields of activity and object are removed.

        Args:
            activity (APActivity): An activity, potentially with an object, both of which can have bto and bcc fields set.

        Returns:
            APActivity: The activity with all bto and bcc fields removed, also from the object.
        """
        if activity.object:
            obj = activity.object
            if isinstance(obj, dict):
                obj = APObject.deserialize(obj)
            if isinstance(obj, APObject):
                obj.bto = None
                obj.bcc = None
                activity.object = obj
        activity.bto = None
        activity.bcc = None
        return activity

    async def deliver_activity(self, activity: APActivity, recipients: list[ObjectId]) -> DeliveryResult:
        """Tries to deliver the given activity server2server, to all listed receipients inboxes.

        Args:
            activity (APActivity): activity, ready to be submitted.
            recipients (list[ObjectId]): List of receipients actor ids (list[URI string]).

        Returns:
            DeliveryResult: Contains all success and failed lists from all individual deliveries, concatenated.
        """
        activity = await ActivityDelivery.remove_hidden_recipients(activity)
        if not activity.id:
            raise DeliveryError("Activity to deliver did not get an id in outbox")

        result = DeliveryResult(success=[], failed=[activity.id], status_code=HTTPStatus.NOT_IMPLEMENTED)

        delivery_queuing = DeliveryQueueType(self.settings.federation.delivery_queuing)
        if delivery_queuing == DeliveryQueueType.Automatic:
            if self.delivery_workflow and self.delivery_queue:
                delivery_queuing = DeliveryQueueType.Worker
            elif self.delivery_queue:
                delivery_queuing = DeliveryQueueType.Queue
            else:
                delivery_queuing = DeliveryQueueType.Immediate

        if delivery_queuing == DeliveryQueueType.Immediate:
            # The inbox is determined by first retrieving the target actor's JSON-LD representation and then looking up the inbox property. If a recipient is a Collection or OrderedCollection, then the server MUST dereference the collection (with the user's credentials) and discover inboxes for each item in the collection. Servers MUST limit the number of layers of indirections through collections which will be performed, which MAY be one.
            # https://www.w3.org/TR/activitypub/#server-to-server-interactions
            recipient_inboxes = await self.resolver.resolve_inboxes(recipients)

            # Servers MUST de-duplicate the final recipient list.
            # https://www.w3.org/TR/activitypub/#delivery
            recipient_inboxes = list(set(recipient_inboxes))

            # also send the activity to all receipients synchronously
            if len(recipient_inboxes) > 0:
                process_queue: list[Awaitable[DeliveryResult]] = []
                for recipient_inbox in recipient_inboxes:
                    process_queue.append(self.deliver_to_url(activity=activity, url=recipient_inbox))

                result = DeliveryResult.create_from_list(await asyncio.gather(*process_queue))
            else:
                result = DeliveryResult(success=[activity.id], status_code=HTTPStatus.CREATED)
        elif delivery_queuing == DeliveryQueueType.Queue or delivery_queuing == DeliveryQueueType.Worker:
            if self.delivery_queue:
                activity_id = await self.delivery_queue.put(activity=activity, recipients=recipients)
                result = DeliveryResult(success=[activity_id], status_code=HTTPStatus.CONTINUE)
            else:
                raise DeliveryError("Delivery queue is not setup")

        if delivery_queuing == DeliveryQueueType.Worker:
            if self.delivery_workflow:
                _ = await self.delivery_workflow()
                result = DeliveryResult(success=[activity.id], status_code=HTTPStatus.CONTINUE)
            else:
                raise DeliveryError("Delivery workflow is not setup")

        return result

    async def deliver_to_local_actor(self, activity: APActivity, actor_id: str, collection: ValidCollection) -> DeliveryResult:
        """Deliver activity to local actor.
        For example, the given url points to inbox or outbox of an actor that exists on the local instance.
        In this case, the delivery can be optimized a lot compared to sending it over the network.

        Args:
            activity (APActivity): to deliver.
            actor_id (str): id of the actor this activity should be delivered to.
            collection (ValidCollection): check to which collection of the actor to send to (only inbox and outbox are supported)

        Returns:
            DeliveryResult: Result of delivery, as if it was sent over the network.
        """
        if not await self.check_is_local(actor_id):
            raise DeliveryError(f"Local delivery to actor '{actor_id}' failed, should only be to the local domain '{self.settings.domain.netloc}'")

        if not self.local_delivery_inbox or not self.local_delivery_outbox:
            raise DeliveryError("Local delivery failed because it is not set up")

        result = DeliveryResult()
        try:
            if collection == ValidCollection.Inbox:
                result = await self.local_delivery_inbox(activity, actor_id)
            elif collection == ValidCollection.Outbox:
                result = await self.local_delivery_outbox(activity)
            else:
                raise DeliveryError(f"Local delivery only works for inbox/outbox right now. Wrong collection choice: '{collection.value}'")

        except ActivityPubException as e:
            self.logger.debug(e.message)
            result.failed.append(actor_id)
            result.error_message = e.user_message
        except Exception as e:
            self.logger.debug(f"Exception while delivering to local actor: {e}")
            result.failed.append(actor_id)
            result.error_message = str(e)

        return result

    async def deliver_to_url(
        self, activity: APActivity, url: ObjectId, retry_count: int = 0, data: UploadFile | None = None, token: SecretStr | None = None
    ) -> DeliveryResult:
        """Deliver activity to a given url (e.g. inbox or outbox of actor).

        Args:
            activity (APActivity): to deliver.
            url (str): url where the activity is to be delivered to.
            retry_count (int, optional): how often the delivery has been tried already. Defaults to 0.
            data (UploadFile, optional): data of a file to upload. Only used in c2s communication (i.e. by client apps).
            token (str, optional): if this is set, the delivery will not use header signing and instead pass the Authorization: Bearer token.

        Returns:
            DeliveryResult: Result of delivery after retries.
        """
        if self.instance and await self.instance.is_blocked_instance(url):
            raise AuthorizationError(f"Instance for url {url} is blocked, cannot deliver")

        parsed_url = urlparse(str(url))
        domain = parsed_url.netloc
        result = DeliveryResult()

        try:
            # Check rate limit
            if not self.rate_limiter.check_rate_limit(domain):
                wait_time = self.rate_limiter.get_wait_time(domain)
                self.logger.info(f"Waiting {int(wait_time)} seconds for rate limit of domain {domain} to stop")
                await asyncio.sleep(wait_time)

            # Prepare headers with host
            headers = PHEDERATION_HEADERS.copy()
            headers["Host"] = domain

            # Get key id of the actor sending this
            actor_id = dereference(activity, key="actor")
            if actor_id is None:
                raise DeliveryError(f"ActivityDelivery: No actor in activity found: {activity}")

            # Add a content hash of the content of the activity.object
            activity = await self.add_hash_to_activity_object(activity)

            # Use standardized JSON serialization
            activity_dict = activity.serialize()
            body_json = to_json_string(activity_dict)

            if data:
                files: dict[str, tuple[None | str, str | BinaryIO, str]] = {
                    "activity_form": (None, body_json, "application/json"),
                    "file": (data.filename, data.file, data.content_type or "text/plain"),
                }
                _ = headers.pop("Content-Type")  # is going to be set by the AsyncClient to multipart
            else:
                files = {}

            if token:
                headers["Authorization"] = f"Bearer {token.get_secret_value()}"
                headers["WWW-Authenticate"] = "Bearer"

            if not token and self.key_manager and self.signature_verifier:
                try:
                    key_id = await self.key_manager.get_key_id_from_actor(actor_id=actor_id)
                except ResolverError as r:
                    self.logger.debug(f"ActivityDelivery: KeyManager throws ResolverError: {r}. Failing silently")
                    result.error_message = "Could not resolve actor."
                    result.status_code = HTTPStatus.CONTINUE
                    result.failed.append(url)  # TODO: should we add to success instead?
                    return result

                # Sign request using HTTPSignatureVerifier
                signed_headers = await self.signature_verifier.sign_request(
                    method="POST",
                    path=parsed_url.path,
                    headers=headers,
                    body_serialized=body_json,
                    key_id=key_id,  # Pass original dict for digest calculation
                )
            else:
                # do not sign headers, use token authentication if available
                signed_headers = headers

            # Make request with pre-serialized JSON or Files
            async def post_files(url: str, client: httpx.AsyncClient):
                return await client.post(url, files=files, headers=signed_headers, timeout=self.settings.federation.delivery_timeout)

            async def post_json(url: str, client: httpx.AsyncClient):
                return await client.post(url, content=body_json, headers=signed_headers, timeout=self.settings.federation.delivery_timeout)

            post_action = post_files if data else post_json

            try:
                # if await self.check_is_local(url):
                #     url_parsed = urlparse(url)
                #     url_local = Url(
                #         scheme="http",
                #         host=self.settings.domain.docker_url,
                #         port=self.settings.domain.docker_port,
                #         fragment=url_parsed.fragment if len(url_parsed.fragment) > 0 else None,
                #         query=url_parsed.query if len(url_parsed.query) > 0 else None,
                #         path=url_parsed.path if len(url_parsed.path) > 0 else None,
                #     ).url
                #     url = url_local
                response = await post_action(url=url, client=self.requests)
            except (httpx.ConnectError, httpx.ConnectTimeout, httpx.ReadTimeout) as e:
                self.logger.debug(f"Error when connecting to url {url}: {type(e).__name__} {e}")
                result.error_message = "Error when connecting to url."
                result.status_code = HTTPStatus.CONTINUE
                result.failed.append(url)
                return result

            response_headers: dict[str, str] = {h.lower(): response.headers[h] for h in response.headers.keys()}

            # Update rate limit state
            if state := self.rate_limiter.update_rate_limit(domain, headers=response_headers):
                self.logger.info(f"Updated rate limit on instance {self.settings.federation.instance_name} for domain {domain}: {state}")

            try:
                response_content: dict[str, Any] = cast(dict[str, Any], json.loads(response.content.decode("utf8")))
            except JSONDecodeError as e:
                raise DecodingError(message=f"JSONDecodeError for response_content '{response.content}'.")

            if "location" in response_headers:
                response_content["Location"] = response_headers["location"]

            if response.status_code in (HTTPStatus.OK, HTTPStatus.CREATED, HTTPStatus.ACCEPTED):
                result.success.append(response_content)
                result.status_code = response.status_code
                return result

            if (
                response.status_code in (HTTPStatus.TOO_MANY_REQUESTS, HTTPStatus.GATEWAY_TIMEOUT)
                and retry_count < self.settings.federation.max_retries
            ):
                # Calculate retry delay
                retry_after = int(response_headers.get("retry-after", self.settings.federation.retry_delay))
                self.logger.warning(f"Retry delivery of activity with id={activity.id} after waiting {retry_after} seconds. Retry_count: {retry_count + 1} / {self.settings.federation.max_retries}...")
                await asyncio.sleep(retry_after)

                # Retry delivery
                return await self.deliver_to_url(activity=activity, url=url, retry_count=retry_count + 1)

            self.logger.warning(
                f"Delivery Error for activity type {activity.type}: url={url}, content={response_content}, headers={response_headers}"
            )
            result.failed.append(url)
            result.status_code = response.status_code
            result.error_message = str(response_content) if len(response_content) > 0 else "No response text received"

        except JSONDecodeError as e:
            self.logger.error(f"Delivery JSONDecodeError: {e}")
            result.failed.append(url)
            result.status_code = HTTPStatus.BAD_REQUEST
            result.error_message = f"Delivery exception: cannot decode."
        except ActivityPubException as e:
            self.logger.error(f"Delivery Exception: {e.message}")
            result.failed.append(url)
            result.status_code = e.status_code
            result.error_message = f"Delivery exception: {e.user_message or 'unknown'}"

        return result

    async def deliver_to_actor_inbox(self, activity: APActivity, actor_id: ObjectId) -> DeliveryResult:
        """Deliver activity to an actor's inbox, i.e., server2server communication."""
        try:
            actor = await self.resolver.resolve_actor(actor_id)

            # Remove all to/cc/bto/bcc/audience fields (also from object), because this is just sent to a single actor.
            activity = await ActivityDelivery.remove_hidden_recipients(activity)
            activity.to = None
            activity.cc = None
            activity.bto = None
            activity.bcc = None
            activity.audience = None

            # Check if the actor is local to the instance. If they are, do not deliver via url, just process it locally.
            if await self.check_is_local(actor_id) and self.local_delivery_inbox is not None:
                return await self.deliver_to_local_actor(activity=activity, collection=ValidCollection.Inbox, actor_id=actor_id)
            else:
                # if this was a locally created activity, do not send this tag to others
                # activity.local = None
                return await self.deliver_to_url(activity=activity, url=actor.inbox)

        except ActivityPubException as e:
            self.logger.debug(f"Discovery failed for actor {actor_id}: {e.message}. Failing silently.")
            return DeliveryResult(failed=[actor_id], error_message="Actor not found.", status_code=HTTPStatus.GONE)
        except Exception as e:
            self.logger.warning(f"Failed to deliver to actor {actor_id} inbox: {e}. Failing silently.")
            return DeliveryResult(failed=[actor_id], error_message=str(e))

    async def add_hash_to_activity_object(self, activity: APActivity) -> APActivity:
        """Compute a SHA256 hash of the activity.object content and add it in the 'hash' field of the object.

        This is only done if the activity is "Create", and if there is not already a hash field set in the object.
        Otherwise, the activity and its object are returned unchanged.

        Args:
            activity (APActivity): Activity with an object attached.

        Returns:
            APActivity: The object with a new 'hash' field, if it was created.
        """
        if (activity.type == ActivityType.CREATE.value or activity.type == ActivityType.UPDATE.value) and activity.object:
            hash: str | None = None
            content: str | None = None
            if isinstance(activity.object, APObject):
                hash = activity.object.hash
                content = activity.object.content
            if hash is None:
                hash = to_sha256(str(content))
            if isinstance(activity.object, dict):
                activity.object["hash"] = hash
            elif isinstance(activity.object, APObject):
                activity.object.hash = hash
        return activity

    async def deliver_to_actor_outbox(self, activity: APActivity, actor_id: ObjectId, token: SecretStr | None = None) -> DeliveryResult:
        """Deliver activity to an actor's outbox.
        This is the main method for client2server communication. Instead of signing the header (which would require access to the private key),
        the client must send a token for authentication.

        Args:
            activity (APActivity): Activity to deliver.
            actor_id (ObjectId): id of actor that is performing the activity.
            token (str | None, optional): Authentication token, if available and necessary. Defaults to None.

        Returns:
            DeliveryResult: Result of the delivery to the outbox. Can contain the id of the created activity, if successful.
        """
        try:
            # always hash the content
            activity = await self.add_hash_to_activity_object(activity)

            # Check if the actor is local to the instance. If they are, do not deliver via url, just process it locally.
            if await self.check_is_local(actor_id) and self.local_delivery_outbox is not None:
                return await self.deliver_to_local_actor(activity=activity, collection=ValidCollection.Outbox, actor_id=actor_id)
            else:
                # this branch is used if the actor is not local to the instance, OR if this method is used from a client.
                # In the latter case, a token is available and can be sent (to the instance).
                actor = await self.resolver.resolve_actor(actor_id)
                return await self.deliver_to_url(activity=activity, url=actor.outbox, token=token)
        except ActivityPubException as e:
            self.logger.error(f"Failed to deliver to actor {actor_id} outbox: {e.message}. Failing silently.")
            return DeliveryResult(failed=[actor_id], error_message=e.user_message)
        except Exception as e:
            self.logger.error(f"Failed to deliver to actor {actor_id} outbox: {str(e)}. Failing silently.")
            return DeliveryResult(failed=[actor_id], error_message=str(e))

    async def deliver_to_federation(self, activity: APActivity):
        """Delivers an activity to the entire network of federated instances.

        This is usually done for APUpdate activities, e.g. key rotation or actors that move instance.
        Sharing is done either to the shared inbox (if available) or the instance actor (as a fallback).

        Args:
            activity (APActivity): An activity to share to the shared inboxes or instance actors of federated instances.
        """
        if not self.instance:
            raise DeliveryError("Delivery does not have an instance object set up. Cannot share to federation")
        if not self.local_delivery_outbox:
            raise DeliveryError("Delivery does not have a local_delivery_outbox set up. Cannot share to federation")

        federated_instance_actors = await self.instance.get_federated_instance_actors(access=AccessType.PRIVATE)

        self.logger.debug(f"Current federated instance actor list: {federated_instance_actors}")

        # set the instances in blind-to field (to hide them from each other)
        activity.bto = federated_instance_actors

        # add hash of the key content to the key
        activity = await self.add_hash_to_activity_object(activity)

        # send through instance actor outbox, to add activity id and store the activity properly
        result = await self.local_delivery_outbox(activity)
        self.logger.info(
            f"Delivering activity {activity.type} with object {activity.object} to {len(federated_instance_actors)} federated instances. Current result: {result}"
        )



"""
Delete activity handler implementation.
"""

from datetime import datetime, timezone
from typing import override

from phederation.models import APTombstone, ValidCollection, dereference
from phederation.models.activities import APActivity
from phederation.models.objects import dereference_or_raise
from phederation.utils import ObjectId, ObjectType
from phederation.utils.base import AccessType
from phederation.utils.exceptions import AuthorizationError, ValidationError

from .base import ActivityHandler


class DeleteHandler(ActivityHandler):
    """Handle Delete activities."""

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Delete activity."""
        # Verify actor permissions
        actor_id = dereference_or_raise(activity, key="actor")
        actor_obj = await self.actor_manager.actor_from_storage(id=actor_id)
        if not actor_obj:
            raise ValidationError("Actor not found")

        # Verify object ownership
        object_id = dereference(activity, key="object")
        if not object_id:
            self.logger.warning("DeleteHandler: Object id not found.")
            raise ValidationError("Cannot find object id")

        if not (await DeleteHandler.check_is_collection(self, actor_id=actor_id, object_id=object_id)):
            stored_object = await self.storage.object.read(id=object_id)
            if not stored_object:
                self.logger.warning("DeleteHandler: Object not found on this instance. Failing silently.")
                return activity

            if stored_object.attributed_to != actor_id:
                raise AuthorizationError("Not authorized to delete this object")

        return activity

    @staticmethod
    async def check_is_collection(handler: ActivityHandler, actor_id: ObjectId, object_id: ObjectId):
        actor_obj = await handler.actor_manager.actor_from_storage(id=actor_id)
        assert actor_obj
        actor_collections_id = dereference(actor_obj, ValidCollection.Collections.value)
        if actor_collections_id and object_id.startswith(actor_collections_id):
            return True
        return False

    async def delete_collection(self, actor_id: ObjectId, actor_collection_id: ObjectId):
        """Remove the entire actor-created collection.

        This removes all items from the collection and also removes the collection link from the actor collections.

        Args:
            actor_id (ObjectId): Id of the actor.
            actor_collection_id (ObjectId): The id of the collection to remove.
        """
        self.logger.info(f"Deleting the actor collection {actor_collection_id}")
        await self.collections.delete_collection(actor_collection_id)

        actor_obj = await self.actor_manager.actor_from_storage(id=actor_id)
        if actor_obj and actor_obj.collections:
            await self.collections.remove_from_collection(actor_obj.collections, items=actor_collection_id, access=AccessType.PRIVATE)

    async def handle_delete(self, activity: APActivity) -> APActivity:
        object_id = dereference_or_raise(activity, key="object")
        actor_id = dereference_or_raise(activity, "actor")

        # check if the object id references an actor collection -> then delete that one
        if await DeleteHandler.check_is_collection(self, actor_id=actor_id, object_id=object_id):
            await self.delete_collection(actor_id=actor_id, actor_collection_id=object_id)
            return activity

        # get object data and then delete it from storage
        obj = await self.storage.object.read(id=object_id)
        if not obj:
            self.logger.warning("DeleteHandler: Object not found on this instance. Failing silently.")
            return activity

        _ = await self.storage.object.delete(id=object_id)

        attributed_to_id = dereference(obj.attributed_to, "id")

        if not attributed_to_id:
            self.logger.warning("DeleteHandler: Object property 'attributed_to' is not set. Failing silently.")
            return activity

        # Create the tombstone at the object_id in the storage
        await self._create_tombstone(object_id=object_id, former_type=ObjectType(obj.type), attributed_to=attributed_to_id)

        # Update collections
        await self._update_collections(object_id=object_id, actor_id=attributed_to_id)

        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Process Delete activity."""
        return await self.handle_delete(activity)

    async def _create_tombstone(self, object_id: ObjectId, former_type: ObjectType, attributed_to: ObjectId) -> None:
        """Create tombstone for deleted object."""
        tombstone = APTombstone(former_type=str(former_type), deleted_at=datetime.now(timezone.utc), id=object_id, attributed_to=attributed_to)
        _ = await self.storage.object.create(tombstone)

    async def _update_collections(self, object_id: ObjectId, actor_id: ObjectId) -> None:
        """Update collections after deletion."""
        # TODO: clean up links
        # remove from the owning actors objects collection, if it exists
        actor = await self.resolver.resolve_actor(actor_id=actor_id)
        if actor.objects:
            # the access type is PRIVATE here, because the original actor tries to delete something they created.
            await self.collections.remove_from_collection(collection_id=actor.objects, items=object_id, access=AccessType.PRIVATE)

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        activity = await self.handle_delete(activity)
        return activity.id



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from http import HTTPStatus
from logging import Logger

from fastapi import APIRouter, FastAPI, HTTPException

from phederation.api.base import BaseAPI
from phederation.federation import ActivityPubServer
from phederation.federation.middleware import ActivityPubMiddleware
from phederation.utils.logging import configure_logger


class BaseRoute(ABC):
    """The base class for Route instances.
    Route objects add routing methods to an app.
    """

    async def initialize(self, app: FastAPI, api: BaseAPI, middleware: ActivityPubMiddleware, server: ActivityPubServer) -> None:
        self.app: FastAPI = app
        self.api: BaseAPI = api
        self.middleware: ActivityPubMiddleware = middleware
        self.server: ActivityPubServer = server
        self.router: APIRouter = APIRouter()
        self.logger: Logger = configure_logger(
            f"{".".join(__name__.split('.')[:-1])}.{type(self).__name__.lower()}", prefix=server.settings.federation.logging_prefix
        )
        self.setup()

    @abstractmethod
    def setup(self) -> None:
        pass


def admin_route(self: BaseRoute):
    def admin_route[**Params, ReturnType](
        function: Callable[Params, Awaitable[ReturnType]],
    ) -> Callable[Params, Awaitable[ReturnType]]:
        """This decorator verifies that the route is running on an instance in admin mode.
        If it fails, a HTTPException(status_code=METHOD_NOT_ALLOWED) is thrown."""

        async def check_is_admin(*args_is_admin: Params.args, **kwargs_is_admin: Params.kwargs) -> ReturnType:
            if not self.server.settings.federation.admin_mode:
                raise HTTPException(
                    status_code=HTTPStatus.METHOD_NOT_ALLOWED, detail={"status": "Error", "description": "Instance is not in admin mode"}
                )
            return await function(*args_is_admin, **kwargs_is_admin)

        return check_is_admin

    return admin_route



"""
client.py
Implements a client to interact with phederation instances.
This is a quality-of-life service to demonstrate and ease new client implementations.
It is also heavily used in testing all functionality of the software.
"""

from http import HTTPStatus
import json
from logging import Logger
from typing import Any, cast

from fastapi.datastructures import QueryParams
import httpx
from pydantic import BaseModel, Field
from pydantic.types import SecretStr

from phederation.cache.dictcache import DictCache
from phederation.federation.ratelimit import RateLimiter
from phederation.models import APCreate, ValidCollection, object_from_type
from phederation.models.activities import APActivity, APMigrate
from phederation.models.actors import APActor
from phederation.models.objects import APDocument, APObject
from phederation.models.proofs import to_sha256
from phederation.security import Authentication
from phederation.utils import ActivityPubBaseWithId, NodeInfo
from phederation.utils.base import (
    InstanceEndpoints,
    ObjectId,
    ResolverResult,
    UrlType,
    actor_id_from_username,
    urljoin,
)
from phederation.utils.exceptions import DeliveryError, ResolverError
from phederation.utils.logging import configure_logger
from phederation.utils.serialization import to_json_string
from phederation.utils.settings import KeycloakSettings, PhedSettings
from phederation.utils.version import PHEDERATION_HEADERS

from .delivery import ActivityDelivery, DeliveryResult
from .discovery import InstanceDiscovery
from .resolver import ActivityPubResolver


class ActivityPubClient(BaseModel):
    """
    ActivityPubClient that can be used in client2server applications to easily setup and communicate with a phederation server using the activitypub protocol.

    Applications can use the 'delivery' and 'resolver' properties of a Client instance to deliver activities, and to resolve information about instances and actors in the federation. The 'host_url' must be set to the phederation server url, and the 'actor_id' must be set to the actor that is operating the client. This actor must be registered on the server.
    """

    settings: PhedSettings = Field(default=..., description="The settings for the client-side phederation setup.")
    instance_actor_id: ObjectId | None = Field(
        default=None,
        description="The id of the instance actor for the host instance. Must be a resolvable URL (as defined by ActivityPub), or None if it should be discovered from the host settings.",
    )
    keycloak_settings: KeycloakSettings | None = Field(description="Settings to reach the keycloak instance for user authentication (OAuth2).")
    username: str | None = Field(default=None, description="Username of the actor to use for login")
    password: SecretStr | None = Field(default=None, description="Password of the actor to use for login")
    access_token: SecretStr | None = Field(default=None, description="Current access_token of the actor to use for authentication")
    _initialized: bool = False

    def initialize(self):
        if not self._initialized:
            self._discovery: InstanceDiscovery = InstanceDiscovery(settings=self.settings)
            self._resolver: ActivityPubResolver = ActivityPubResolver(settings=self.settings)
            self._rate_limiter: RateLimiter = RateLimiter(cache=DictCache(), settings=self.settings)
            self._delivery: ActivityDelivery = ActivityDelivery(
                discovery=self._discovery,
                resolver=self._resolver,
                rate_limiter=self._rate_limiter,
                settings=self.settings,
            )
            self._logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix + "_client")
            if self.keycloak_settings is None:
                self._logger.warning("Client does not have keycloak settings set. Will not be able to authenticate")
            self._authentication: Authentication = Authentication(settings=self.settings, keycloak_settings_client=self.keycloak_settings)
        self._initialized = True

    async def resolve_multiple(self, urls: list[ObjectId]):
        """Tries to resolve multiple objects given by their id.
        This will use the /resolve endpoint of the instance.
        The objects in the list can be of different types.
        The object ids must be for the instance connected to this client (because this is using the instance resolve endpoint)

        Args:
            urls (list[ObjectId]): Ids of objects to resolve.

        Returns:
        list[APObject | DataIntegrityProof | NodeInfo]: list of resolved objects. If an object id cannot be resolved, it is not included. The list is not necessarily in the same order as the given urls, but the objects have their own ids, so can be identified again.
        """
        # TODO: allow to resolve from multiple instance endpoints
        # get resolve endpoint from instance
        if not self._initialized:
            raise ResolverError("Must initialize the client first")
        instance_info = await self._discovery.discover_instance(self.settings.domain.base_url)
        resolve_endpoint: ObjectId | None = None
        if isinstance(instance_info.endpoints, InstanceEndpoints):
            resolve_endpoint = instance_info.endpoints.resolve
        elif "resolve" in list(instance_info.endpoints.keys()):
            resolve_endpoint = instance_info.endpoints.get("resolve", None)
        if resolve_endpoint:
            try:
                headers = await self._headers(with_token=True)
                urls_encoded = json.dumps(urls)
                response = await self.delivery.requests.post(resolve_endpoint, headers=headers, content=urls_encoded)
                result_bytes = await response.aread()
                result_json = cast(list[dict[str, Any]], json.loads(result_bytes.decode(encoding="utf-8")))
                result_objects = [object_from_type(obj_dict) for obj_dict in result_json]
                result_objects_with_id = [obj for obj in result_objects if isinstance(obj, ActivityPubBaseWithId)]
                return result_objects_with_id
            except Exception as e:
                raise ResolverError(f"Could not resolve from /resolve endpoint; error: {e}")
        raise ResolverError(f"The instance does not have a /resolve endpoint")

    @property
    def actor_id(self):
        """The id of the actor who is using this client.
        A resolvable URL (as defined by ActivityPub), or None if no specific actor is using it."""
        if self.username:
            return actor_id_from_username(self.settings.domain.base_url, username=self.username)
        else:
            raise ValueError("Actor_id not available in ActivityPubClient: field 'username' is not set")

    def get_token(self) -> SecretStr | None:
        if self.access_token:
            return self.access_token
        if not self._authentication or not self.username or not self.password:
            token = None
        else:
            token = self._authentication.authenticate_user(username=self.username, password=self.password)
            self.access_token = token
        return token

    async def deliver_activity(self, activity: APActivity):
        if not self.actor_id:
            raise DeliveryError("Could not deliver activity, actor_id is not set")
        token = self.get_token()
        obj = activity.object
        if self.settings.media.create_file_hash and isinstance(obj, APObject):
            obj.hash = to_sha256(obj.content or "")

        result: DeliveryResult | None = None
        for k in range(self.settings.federation.max_retries):
            try:
                result = await self.delivery.deliver_to_actor_outbox(activity, actor_id=self.actor_id, token=token)
                break
            except httpx.ReadTimeout:
                self._logger.warning(f"Http ReadTimeout when delivering activity, trying again {k+1}/{self.settings.federation.max_retries}...")
        if not result:
            result = DeliveryResult(
                success=[],
                failed=[self.actor_id],
                status_code=HTTPStatus.BAD_GATEWAY,
                error_message=f"Httpx ReadTimeout errors, tried {self.settings.federation.max_retries} times.",
            )
        return result

    async def deliver_maintenance(self, activity: APActivity):
        token = self.get_token()
        obj = activity.object
        if self.settings.media.create_file_hash and isinstance(obj, APObject):
            obj.hash = to_sha256(obj.content or "")

        maintenance_route = urljoin(self.settings.domain.admin_url, UrlType.Maintenance.value)
        result = await self.delivery.deliver_to_url(activity, url=maintenance_route, token=token)
        return result

    async def resolve_nodeinfo(self) -> NodeInfo | None:
        base_url = self.settings.domain.base_url
        nodeinfo = await self._discovery.discover_nodeinfo(base_url=base_url)
        return nodeinfo

    async def register_user(self):
        """Registers the user associated with this client on the instance keycloak.

        Note that this right now only works with the MockKeycloak, not an actual keycloak instance.

        Raises:
            DeliveryError: If the client is not set up properly.

        Returns:
            fastapi.Response: the response from the keycloak after registration.
        """
        if not self.actor_id or not self.keycloak_settings:
            raise DeliveryError("Could not register user, actor_id or keycloak settings not defined")
        if not self.password or not self.username:
            raise DeliveryError("Could not register user, username or password not defined")

        url = urljoin(self.settings.domain.keycloak_hostname, f"/realms/{self.keycloak_settings.keycloak_realm}/protocol/openid-connect/register")
        headers = await self._headers(with_token=False)
        password_string = self.password.get_secret_value()
        params = {"username": self.username, "password": password_string, "redirect_uri": self.settings.domain.base_url}
        result = await self.delivery.requests.put(url=url, params=QueryParams(params), headers=headers)
        return result

    async def create_account(
        self,
    ):
        if not self.actor_id or not self.username:
            raise DeliveryError("Could not create account, actor_id or username is not set")
        actor_id = actor_id_from_username(self.settings.domain.base_url, username=self.username)
        inbox_url = urljoin(actor_id, ValidCollection.Inbox.value)
        outbox_url = urljoin(actor_id, ValidCollection.Outbox.value)
        activity = APCreate(actor=self.actor_id, object=APActor(preferred_username=self.username, inbox=inbox_url, outbox=outbox_url))
        url = urljoin(self.settings.domain.base_url, ValidCollection.Users.value)
        headers = await self._headers(with_token=True)
        body_serialized = to_json_string(activity.serialize())
        self._logger.info(f"Sending user acount registration for user {self.username} to instance url {url}")
        result = await self.delivery.requests.put(url=url, content=body_serialized, headers=headers)
        return result

    async def _headers(self, with_token: bool = True):
        headers = PHEDERATION_HEADERS
        headers["Host"] = self.settings.domain.url
        if with_token:
            token = self.get_token()
            if token:
                headers["Authorization"] = f"Bearer {token.get_secret_value()}"
                headers["WWW-Authenticate"] = "Bearer"
        return headers

    @property
    def delivery(self):
        self.initialize()
        return self._delivery

    @property
    def resolver(self):
        self.initialize()
        return self._resolver

    async def discover_instance_actor(self):
        if self.instance_actor_id is None:
            host_instance_url = self.settings.domain.base_url
            actor = await self._discovery.discover_instance_actor(host_instance_url)
            if actor and "id" in actor:
                self.instance_actor_id = actor.get("id", None)
        return self.instance_actor_id

    async def download_user_data(self):
        if not self.username:
            raise ResolverError("Username is not set in the client, cannot download user data")
        # send the "assemble user data" command
        create_userdata = APMigrate(actor=self.actor_id)
        result = await self.deliver_activity(activity=create_userdata)
        activity_id = result.get_success_id_or_raise()

        # retrieve the activity to get the file url
        token = self.get_token()
        activity = await self.resolver.resolve_activity(activity_id=activity_id, access_token=token)
        file_descriptor_document = await self.resolver.resolve_object(activity.object)
        file_urls = file_descriptor_document.url

        if not file_urls:
            response = ResolverResult(
                content=file_descriptor_document.id, status_code=None, error_message="Url is not contained in the user data document"
            )
            raise ResolverError("Could not resolve any user data file urls", response=response)

        # get the actor media endpoint and download the data
        files: list[bytes] = []
        for file_url in file_urls:
            resolved_file = b"".join([line async for line in await self.resolver.resolve_media(file_id=file_url, access_token=token)])
            files.append(resolved_file)
        return files

    async def user_migration_initialize(self, client_for_new_instance: "ActivityPubClient"):
        activity = APMigrate(actor=self.actor_id, actor_to=client_for_new_instance.actor_id, object=None)
        result = await self.deliver_activity(activity)
        return result

    async def user_migration_status(self, migration_activity_id: ObjectId):
        token = self.get_token()
        activity = await self.resolver.resolve_activity(migration_activity_id, access_token=token)
        if activity.object:
            obj = await self.resolver.resolve_object(activity.object, access_token=token)
            return cast(APDocument, obj)



"""Integration into a web framework.

Defines the interface to a proper web endpoint and API, here FastAPI.
"""


from .base import BaseIntegration
from .integration import FastAPIIntegration

__all__ = [
    "BaseIntegration", "FastAPIIntegration"
]


"""
Collection handling implementation.
"""

import asyncio
import datetime
import math
from typing import override

from phederation.collections.collections import CollectionManager
from phederation.models import dereference
from phederation.models.collections import APOrderedCollectionPage
from phederation.models.objects import APCollectionItem
from phederation.storage.base import StorageBackend
from phederation.utils import ObjectId
from phederation.utils.base import AccessType


class CollectionManagerNoDB(CollectionManager):
    """Manages ActivityPub collections, without storing any collection in the database (instead using pagination from StorageBackend)."""

    def __init__(self, storage: StorageBackend, page_size: int = 20):
        super(CollectionManagerNoDB, self).__init__(storage=storage, page_size=page_size)

    @override
    async def initialize(self):
        pass

    @override
    async def create_collection(
        self,
        collection_id: ObjectId,
        items: list[ObjectId] | None = None,
        attributed_to: ObjectId | None = None,
        access: AccessType = AccessType.PUBLIC,
    ) -> ObjectId:
        """
        Create a new collection.
        In the nodb setting, this method does nothing if no items are set - a collection does not exist without items.

        Args:
            collection_id: Collection id to create
            items: Initial collection items
            attributed_to: the actor or object this collection is attributed to.
            access (AccessType): access type of the individual items added. Default: PUBLIC

        Returns:
            Collection ID
        """
        if items:
            await self.add_to_collection(collection_id=collection_id, items=items, access=access)
        return collection_id

    @override
    async def pages(self, collection_id: ObjectId, attributed_to: ObjectId | None = None, access: AccessType = AccessType.PUBLIC):
        """Iterate over all pages of the collection.

        Args:
            collection_id (ObjectId): id of the collection. Can include a page, but will always start with the first item.

        Raises:
            CollectionError: If a part of the collection cannot be found.

        Yields:
            APOrderedCollectionPage: pages of the collection, one by one.
        """
        total_items: int = await self.storage.collection_items.count(collection_id=collection_id, access=access)
        n_pages: int = max(1, math.ceil(total_items / float(self.storage.settings.storage.page_size)))
        page_number = 1
        async for page_items in self.storage.paginate(collection_id=collection_id, access=access):
            yield await self.create_new_page(
                collection_id=collection_id,
                page=page_number,
                n_pages=n_pages,
                total_items=total_items,
                attributed_to=attributed_to,
                items=[i.stored_id for i in page_items] if page_items else [],
            )
            page_number += 1

    @override
    async def contains(self, collection_id: ObjectId, item: ObjectId | None, access: AccessType = AccessType.PUBLIC) -> bool:
        """Check if the collection contains the given item.

        Args:
            collection_id (ObjectId): id of the collection
            item (ObjectId | None): item to check. Always returns False if item is None.

        Returns:
            bool: True if contained, False otherwise.
        """
        if not item:
            return False
        content = await self.storage.collection_items.select(collection_id=collection_id, stored_ids=[item], access=access)
        return len(content) > 0

    @override
    async def add_to_collection(
        self, collection_id: ObjectId, items: ObjectId | list[ObjectId], access: AccessType = AccessType.PUBLIC, add_only_once: bool = False
    ) -> None:
        """
        Add items to collection.

        Args:
            collection_id: Collection ID
            items: Item(s) to add
        """
        if isinstance(items, str):
            items = [items]
        if len(items) == 0:
            return
        collection_items = [
            APCollectionItem(
                collection_id=collection_id,
                stored_id=item,
                visibility=access.value,
                created_at=datetime.datetime.now(datetime.timezone.utc),
                id=f"{collection_id}_{item}" if add_only_once else None,
            )
            for item in items
        ]
        if len(collection_items) == 1:
            _ = await self.storage.collection_items.create(collection_items[0], raise_if_exists=False)
        else:
            _ = await self.storage.collection_items.create_many(collection_items)

    @override
    async def remove_from_collection(self, collection_id: ObjectId, items: ObjectId | list[ObjectId], access: AccessType = AccessType.PUBLIC) -> None:
        """
        Remove items from collection.
        This deletes the entire collection, including all pages, and reforms it - meaning it is not very efficient.

        Args:
            collection_id: Collection ID
            items: Item(s) to remove
        """
        # only use lists
        if isinstance(items, str):
            items = [items]
        if len(items) == 0:
            return

        # get all accessible items of the collection, remove the list
        item_ids = await self.storage.collection_items.select(collection_id=collection_id, stored_ids=items, access=access)
        delete_tasks = [self.storage.collection_items.delete(id=item.id) for item in item_ids if item.id]
        _ = await asyncio.gather(*delete_tasks)

    @override
    async def delete_collection(self, collection_id: ObjectId) -> None:
        """Deletes the entire collection (with AccessType.PRIVATE), including all pages, from storage.
        Items within the collection are not affected.

        Args:
            collection_id (ObjectId): The collection to remove.
        """
        items = await self.get_collection_items(collection_id=collection_id, access=AccessType.PRIVATE)
        await self.remove_from_collection(collection_id=collection_id, items=items, access=AccessType.PRIVATE)

    @override
    async def collection_exists(self, collection_id: ObjectId, page: int | None = None, access: AccessType = AccessType.PUBLIC) -> bool:
        """Check if the collection exists.

        This will check how many items are accessible by access in the collection.
        If none are there, it will return False.

        Args:
            collection_id (ObjectId): id of the collection to check.
            page (int | None, optional): Optional, only check if the given page exists. Defaults to None.
            access: which items to count

        Returns:
            bool: True if collection has more than zero items, or False.
        """
        n_items = await self.storage.collection_items.count(collection_id=collection_id, access=access)
        return n_items > 0

    @override
    async def get_collection_page(
        self, collection_id: ObjectId, attributed_to: ObjectId | None = None, page: int = 1, access: AccessType = AccessType.PUBLIC
    ) -> APOrderedCollectionPage:
        """
        Get collection page.

        Args:
            collection_id: Collection ID
            page: Page number

        Returns:
            Collection page
        """
        total_items: int = await self.storage.collection_items.count(collection_id=collection_id, access=access)
        n_pages: int = max(1, math.ceil(total_items / float(self.storage.settings.storage.page_size)))
        page_items = await self.storage.get_collection_page_items(collection_id=collection_id, page=page, access=access)
        return await self.create_new_page(
            collection_id=collection_id,
            page=page,
            n_pages=n_pages,
            total_items=total_items,
            attributed_to=attributed_to,
            items=[i.stored_id for i in page_items] if page_items else [],
        )

    @override
    async def get_collection_items(self, collection_id: ObjectId, access: AccessType = AccessType.PUBLIC) -> list[ObjectId]:
        """Returns all item ids of all pages in the entire collection, not paginated.

        Args:
            collection_id (ObjectId): Id of the collection

        Returns:
            list[ObjectId]: All item ids.
        """
        result: list[ObjectId] = []
        async for page in self.pages(collection_id=collection_id, attributed_to=None, access=access):
            items = page.items or page.ordered_items or []
            dereferenced = [dereference(item, "id") for item in items]
            result.extend([id for id in dereferenced if id])
        return result



"""
Enhanced key management with rotation support.
"""

import asyncio
import base64
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
import enum
from logging import Logger
from uuid import uuid4

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.types import (
    PrivateKeyTypes,
    PublicKeyTypes,
)

from phederation.cache import BaseCache
from phederation.cache.base import WithCache, with_cache
from phederation.federation.resolver import ActivityPubResolver
from phederation.models import APPublicKey, DataIntegrityProof
from phederation.models.keys import APPrivateKey
from phederation.storage.base import StorageBackend
from phederation.utils import ObjectId
from phederation.utils.base import AccessType, UrlType, assemble_id_url
from phederation.utils.exceptions import (
    KeyManagementError,
    ResolverError,
    SecurityError,
    catch_exceptions,
)
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings


class KeyType(str, enum.Enum):
    RSA = "rsa-sha256-key"
    Fernet = "fernet-key"


@dataclass
class KeyPair:
    key_id: ObjectId
    private_key: RSAPrivateKey | None = None
    public_key: RSAPublicKey | None = None
    created_at: datetime | None = None
    expires_at: datetime | None = None


class KeyManager(WithCache):
    """
    Enhanced key management with rotation.

    This class handles private/public RSA key generation.
    Keys are stored on disk, and stored in an "active_keys" dictionary for easy access.
    Keys expire after a set amount of days (specified in the KeyRotation settings).
    The key_id is an url that points to the location where information about the public key can be gathered.

    Note that keys are not associated to any user or the instance, this is handled outside of the class.
    """

    def __init__(
        self,
        settings: PhedSettings,
        resolver: ActivityPubResolver,
        storage: StorageBackend,
        cache: BaseCache | None = None,
    ):
        """
        Creates a new KeyManager.

        Args:
            settings (PhedSettings): to read the domain url of the current instance, and key rotation settings.
            resolver (ActivityPubResolver): resolver object to resolve keys over the network.
            storage (StorageBackend): storage object to store and retrieve keys in a database.
            cache (BaseCache | None, optional): Cache to store remote public keys. Defaults to None.
        """
        self.settings: PhedSettings = settings
        self.resolver: ActivityPubResolver = resolver
        self.storage: StorageBackend = storage
        self.cache: BaseCache | None = cache
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

    async def initialize(self) -> None:
        """Initialize key manager."""
        pass

    async def resolve_current_key(self, key_ids: list[ObjectId] | ObjectId | None):
        """From the given list of key ids, return the most recently created one.

        This is useful to get the most recent key of an APActor, especially after key rotation.

        Args:
            key_ids (list[ObjectId]): List of key ids.
        """
        if not key_ids:
            return None
        if isinstance(key_ids, str):
            key_ids = [key_ids]

        keys = await asyncio.gather(*[self.resolve_key(key_id) for key_id in key_ids])

        def sort_key_by_creation_date(key: APPublicKey):
            return key.created_at or datetime.now(timezone.utc)

        keys.sort(key=sort_key_by_creation_date, reverse=True)
        if len(keys) > 0:
            return keys[0]
        else:
            return None

    @with_cache(prefix=__name__ + "get_key_id_from_actor")
    async def get_key_id_from_actor(self, actor_id: ObjectId) -> ObjectId:
        """Obtain the valid id of the given actor.

        Args:
            actor_id (ObjectId): id of the actor

        Raises:
            ResolverError: If the actor or the publicKey property could not be resolved.

        Returns:
            str: the local id (not the URL, but the integer!) of the key of the actor.
        """
        actor = await self.resolver.resolve_actor(actor_id=actor_id)
        if not actor:
            self.logger.warning(f"Failed to read actor with id {actor_id}")
            raise ResolverError(f"Could not read actor")
        all_keys = actor.public_key_as_ids()
        self.logger.debug(f"Resolved actor {actor_id}, get current key from keys {all_keys}...")
        public_key = await self.resolve_current_key(all_keys)

        if public_key and public_key.id:
            return public_key.id
        else:
            self.logger.warning(f"Failed to get 'public_key' property from actor with id {actor_id}; dict={actor.serialize()}")
            raise ResolverError(f"Could not resolve actor property 'public_key / publicKey'. Maybe not a local actor")

    @catch_exceptions(KeyManagementError, "Fernet key generation failed")
    async def generate_fernet_key(self, actor_id: ObjectId | None) -> tuple[ObjectId, bytes]:
        """Generates a new Fernet key and stores it to storage.

        Note that this key is stored unencrypted in the storage database.
        Anyone with direct access to that database can access all information that is encrypted with this key.

        Usually only the messages to and from the temporalio backend are encrypted with this key, so
        this should not be an issue.

        Args:
            actor_id: The id of the actor this key is associated with.

        Returns:
            ObjectId, bytes: tuple of (The key id, the raw Fernet key).
        """
        # generate the key
        key_bytes = Fernet.generate_key()

        public_pem = "NO PUBLIC KEY PEM"
        private_pem = base64.encodebytes(key_bytes).decode()

        # Set validity period
        created_at = datetime.now(timezone.utc)
        expires_at = created_at + timedelta(seconds=self.settings.security.key_rotation_period)

        # Generate key ID (for HTTP use)
        key_primary = str(uuid4())
        key_id = assemble_id_url(type=UrlType.Keys, base_url=self.settings.domain.base_url, primary=key_primary)

        key = APPrivateKey(
            id=key_id,
            attributed_to=actor_id,
            media_type=KeyType.Fernet.value,
            public_key_pem=public_pem,
            private_key_pem=private_pem,
            created_at=created_at,
            expires_at=expires_at,
            visibility=AccessType.PRIVATE.value,
        )
        # TODO: also add this key to the list of keys of the given actor, so that it can be replaced during key rotation
        key_id = await self.storage.key.create(data=key)
        return key_id, key_bytes

    @catch_exceptions(KeyManagementError, "RSA key generation failed")
    async def generate_key_pair(self, actor_id: ObjectId | None = None) -> KeyPair:
        """Generates a new RSA key pair and stores it to storage.

        Parameters:
            actor_id (ObjectId): if the key is for an actor, provide the id here. Default: None

        Raises:
            KeyManagementError: If the key generation fails.

        Returns:
            KeyPair: The generated keys (private and public key).
        """
        self.logger.debug("Generating new key pair")
        # Generate keys
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=self.settings.security.key_size)
        public_key = private_key.public_key()

        # Set validity period
        created_at = datetime.now(timezone.utc)
        expires_at = created_at + timedelta(seconds=self.settings.security.key_rotation_period)

        # Generate key ID (for HTTP use)
        key_primary = str(uuid4())
        key_id = assemble_id_url(type=UrlType.Keys, base_url=self.settings.domain.base_url, primary=key_primary)

        self.logger.debug(f"Generated key ID: {key_id}")

        # Create key pair
        key_pair = KeyPair(private_key=private_key, public_key=public_key, created_at=created_at, expires_at=expires_at, key_id=key_id)

        # Save keys in storage
        await self._save_RSA_key_pair(key_pair, actor_id=actor_id)
        self.logger.debug("Saved key pair to storage")

        return key_pair

    @with_cache(__name__ + "_resolve_key")
    async def resolve_key(self, key_id: ObjectId) -> APPrivateKey | APPublicKey:
        """Get the APPrivateKey for a given key_id.
        If only the public key is available, return an APPRivateKey or APPublicKey.

        Args:
            key_id (str): id number of the given key

        Returns:
            APPrivateKey | APPublicKey: The key data associated to the id.
        """
        # check if the key is available locally
        key_obj = await self.storage.key.read(id=key_id)

        # if not, resolve over network
        if not key_obj or not key_obj.public_key_pem:
            if self.resolver.is_local(url=key_id):
                raise ResolverError(f"Key {key_id} could not be found locally, but has a local url")
            self.logger.debug(f"Key {key_id} is not local, resolving public key over network")
            # given an actor, check if we have its public key already. If not, resolve it and add it to the key manager.
            key_obj = await self.resolver.resolve_object(object_id=key_id)
            if isinstance(key_obj, APPrivateKey) and key_obj.public_key_pem:
                self.logger.debug(f"Key loaded successfully: keyId={key_id}")
            else:
                self.logger.debug(f"KeyManager: object_dict={key_obj}, not resolved for key with id={key_id}.")
                raise ResolverError(
                    f"KeyManager: object_dict={key_obj}, not resolved for key with id={key_id}.", user_facing_message="Could not resolve key"
                )

        if not key_obj.private_key_pem:
            key_obj = APPublicKey.deserialize(key_obj.serialize())
            if not isinstance(key_obj, APPublicKey):
                raise ResolverError(f"Could not resolve public key.")
        return key_obj

    @with_cache(__name__ + "get_RSA_key_pair")
    async def get_RSA_key_pair(self, key_id: ObjectId) -> KeyPair:
        """Get the RSA KeyPair for a given key_id, including the private key.

        Args:
            key_id (str): id number of the given key

        Returns:
            KeyPair: The KeyPair associated to the id.
        """
        key = await self.resolve_key(key_id=key_id)

        public_key: PublicKeyTypes = serialization.load_pem_public_key(str(key.public_key_pem).encode())
        private_key: PrivateKeyTypes | None = None
        if isinstance(key, APPrivateKey) and key.private_key_pem:
            private_key = serialization.load_pem_private_key(str(key.private_key_pem).encode(), password=None)

        if not isinstance(public_key, RSAPublicKey):
            raise KeyManagementError("public key PEM did not resolve to RSAPublicKey")
        if private_key and not isinstance(private_key, RSAPrivateKey):
            raise KeyManagementError("private key PEM did not resolve to RSAPublicKey")

        key_pair = KeyPair(key_id=key_id, private_key=private_key, public_key=public_key, created_at=key.created_at, expires_at=key.expires_at)

        return key_pair

    async def get_public_key_APObject(self, key_id: ObjectId) -> APPublicKey:
        key = await self.resolve_key(key_id)
        if isinstance(key, APPrivateKey):
            key.private_key_pem = None
            public_key = APPublicKey.deserialize(key.serialize())
            if not isinstance(public_key, APPublicKey):
                raise KeyManagementError(f"Could not deserialize public key from private key, key_id={key_id}")
        else:
            public_key = key
        return public_key

    async def get_public_key(self, key_id: ObjectId) -> RSAPublicKey | None:
        key = await self.get_RSA_key_pair(key_id=key_id)
        return key.public_key

    async def get_public_key_pem(self, key_id: ObjectId) -> str | None:
        """Get the public key in PEM format for a key_id.

        Args:
            key_id (str): id number of the given key

        Returns:
            str: A string representation of the public PEM key file.
        """
        key = await self.get_public_key(key_id=key_id)
        if key:
            return key.public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo).decode("utf-8")

    async def verify_key(self, key_id: ObjectId) -> bool:
        """Verify a key's validity.
        Checks if the given key is in the active keys and not expired yet.
        Does not check external keys yet (domain is not used).

        Args:
            key_id (str): key id (URL) for the key to verify.

        Returns:
            bool: True (valid key) or False (invalid).
        """
        try:
            # Check if key is one of our active keys
            key = await self.storage.key.read(id=key_id)
            if key and key.expires_at:
                return datetime.now(timezone.utc) <= key.expires_at

            # TODO: For external keys, verify with their server implementation for external key verification
            return False

        except Exception as e:
            self.logger.error(f"Key verification failed: {e}")
            return False

    async def rotate_key(self, key: APPublicKey | ObjectId, actor_id: ObjectId | None) -> ObjectId:
        """Rotates the given key.
        If the actor id is set, also replaces the public key id for this actor with the new key.

        Args:
            key (APPublicKey): Key to replace.
            actor_id (ObjectId | None): If given, will check if the key is associated to this actor and also replace it there.

        Returns:
            ObjectId: the id of the new key, or none if it could not be generated.
        """
        # resolve key if only given an id
        if isinstance(key, str):
            key = await self.resolve_key(key_id=key)

        key_id_old = key.id
        if not key_id_old:
            raise KeyManagementError("Key id of old key before rotation should not be None")

        # replace key in actor if available
        if actor_id:
            key_id_from_actor = await self.get_key_id_from_actor(actor_id=actor_id)
            if key_id_from_actor == key.id:
                # replace the public key of the actor
                actor = await self.storage.actor.read(id=actor_id)
                if actor:
                    # generate new key
                    if key.media_type == KeyType.Fernet.value:
                        key_id_new, _ = await self.generate_fernet_key(actor_id=actor_id)
                    else:
                        key_pair = await self.generate_key_pair(actor_id=actor_id)
                        key_id_new = key_pair.key_id
                    # add new key to actor key list and update
                    actor.add_public_key(key_id_new)
                    _ = await self.storage.actor.upsert(id=actor_id, data=actor)
                    return key_id_new
                else:
                    raise KeyManagementError(f"Failed to read actor {actor_id} to replace the expired public key")
            else:
                raise KeyManagementError(f"Key id from {actor_id} is not the same as the one that should be replaced")
        else:
            raise KeyManagementError("Actor id for key is None")

    @catch_exceptions(KeyManagementError, "Failed to save key pair")
    async def _save_RSA_key_pair(self, key_pair: KeyPair, actor_id: ObjectId | None = None) -> None:
        """Save key pair to storage."""
        if not key_pair.private_key or not key_pair.public_key:
            raise KeyManagementError("Public or private key missing")

        private_pem = key_pair.private_key.private_bytes(
            encoding=serialization.Encoding.PEM, format=serialization.PrivateFormat.PKCS8, encryption_algorithm=serialization.NoEncryption()
        ).decode(encoding="utf-8")

        public_pem = key_pair.public_key.public_bytes(
            encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode(encoding="utf-8")

        key = APPrivateKey(
            id=key_pair.key_id,
            attributed_to=actor_id,
            public_key_pem=public_pem,
            private_key_pem=private_pem,
            media_type=KeyType.RSA.value,
            created_at=key_pair.created_at,
            expires_at=key_pair.expires_at,
        )
        await self.store_key(key)
    
    async def store_key(self, key: APPrivateKey | APPublicKey):
        if not isinstance(key, APPrivateKey):
            key = key = APPrivateKey(
                id=key.id,
                attributed_to=key.attributed_to,
                public_key_pem=key.public_key_pem,
                private_key_pem=None,
                media_type=key.media_type or KeyType.RSA.value,
                created_at=key.created_at,
                expires_at=key.expires_at,
            )
        if key.id and not await self.storage.key.read(id=key.id):
            _ = await self.storage.key.create(data=key)

    async def proof_string(self, string_to_sign: str, private_key_id: ObjectId | None) -> DataIntegrityProof:
        """Creates a DataIntegrityProof for the given string, using the private key with the given id.

        The cryptosuite is fixed to 'RSA', because the method can only deal with RSA private keys for now.
        The proof_purpose is always 'assertionMethod'.

        Args:
            string_to_sign (str): The string to create the proof for.
            private_key_id (ObjectId | None): Id of the private key to use for signing.

        Raises:
            SecurityError: If the private key cannot be found.

        Returns:
            DataIntegrityProof: An RSA proof for the given string.
        """
        if not private_key_id:
            raise SecurityError(f"Key with id {private_key_id} not found")
        proof = DataIntegrityProof(
            type="DataIntegrityProof",
            cryptosuite="RSA",
            proof_purpose="assertionMethod",
            verification_method=private_key_id,
        )
        key_pair = await self.get_RSA_key_pair(key_id=private_key_id)
        if key_pair.private_key is None:
            raise SecurityError(f"Private key {private_key_id} not found")

        proof = proof.sign(string_to_sign, key=key_pair.private_key)
        proof_id = await self.storage.proof.create(data=proof)
        proof.id = proof_id
        return proof

    async def close(self) -> None:
        """Clean up resources."""
        pass



"""
Phederation settings.
"""

import enum
from typing import ClassVar, override
from urllib.parse import urlparse

from pydantic import BaseModel, Field, SecretStr
from pydantic_core import Url
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)

from phederation.utils.base import ObjectId, RateLimit


class StorageProviderType(str, enum.Enum):
    Sqlite = "sqlite"
    Postgresql = "postgres"
    Mongodb = "mongodb"


class DeliveryQueueType(str, enum.Enum):
    """Type of delivery for messages."""

    Immediate = "immediate"
    """Immediately sends the message, does not use temporalio."""

    Queue = "queue"
    """Uses a periodically scheduled internal task to poll the message queue and delivers the message asynchronously."""

    Worker = "worker"
    """Tries to connect to the temporalio workflow and delivers the message asynchronously."""

    Automatic = "auto"
    """Chooses between 'Immediate' (lowest priority), 'Queue' (second priority), and 'Worker' (highest priority) depending on the availability of the queue and temporalio."""


class KeycloakSettings(BaseSettings):
    """Separate settings model for the keycloak connection.

    This is only used in `Authentication` and `Authorization` classes.
    It is not part of the `PhedSettings` that are passed to all main classes.
    """

    keycloak_realm: str = Field(default="phederation", alias="KEYCLOAK_REALM")
    keycloak_client_id: str = Field(default="phederation", alias="KEYCLOAK_CLIENT_ID")
    keycloak_client_secret: SecretStr = Field(default=SecretStr("<client secret not set>"), alias="KEYCLOAK_CLIENT_SECRET")

    model_config: ClassVar[SettingsConfigDict] = SettingsConfigDict(
        env_file=".env", env_file_encoding="utf-8", extra="allow", arbitrary_types_allowed=True
    )


class DomainSettings(BaseModel):
    """Sets the instance-specific URL and ports, including links within a docker environment."""

    url: str = Field(default="localhost", description="The url this instance can be accessed by from outside.")
    port: int | None = Field(default=None, description="Port this instance can be accessed by from outside. Default None, implies port 80.")
    protocol: str = "https"
    docker_url: str | None = Field(
        default=None, description="The url used internally in a docker environment, to access the API without routing the through external network"
    )
    docker_port: int | None = Field(
        default=None, description="The port used internally in a docker environment, to access the API without routing the through external network"
    )
    admin_port: int | None = Field(default=9082, description="The port opened and used for admin mode. It should not be exposed to the web.")

    keycloak_hostname: str = Field(
        default="http://localhost:8070", description="The address (including schema and port) that keycloak is available at"
    )
    keycloak_docker_hostname: str = Field(
        default="http://keycloak:8070",
        description="The address (including schema and port) that keycloak is available at within a docker environment",
    )
    keycloak_port: int = Field(default=8070, description="The port that keycloak is available at")

    @property
    def base_url(self):
        """Constructs an url with "protocol://url:port", i.e. the external schema and netloc of the instance."""
        return Url.build(scheme=self.protocol, host=self.url, port=self.port).__str__()

    @property
    def admin_url(self):
        """Constructs the url of the admin mode, which is always "http://127.0.0.1:<admin_port>"."""
        return Url.build(scheme="http", host="127.0.0.1", port=self.admin_port).__str__()

    @property
    def netloc(self):
        """Constructs an url without the schema, just "url:port", i.e. the external netloc of the instance."""
        base_parsed = urlparse(self.base_url)
        return base_parsed.netloc

    def is_local(self, id: ObjectId) -> bool:
        """Checks if the given id is from an object local to this instance."""
        return id.startswith(self.base_url)


class RatelimiterSettings(BaseModel):
    # TODO: implement rate limit strategy
    strategy: str = Field(default="SLIDING_WINDOW", description="Rate limiting strategy, not implemented")

    default_policy: RateLimit = Field(default=RateLimit(requests=60, period=60), description="Default policy for a new IP, if no other policy applies")
    policies: dict[str, RateLimit] = Field(default={}, description="Special policies for certain domains")


class DatabaseSettings(BaseModel):
    """Settings for the database connection."""

    name: str = Field(default="phederation", description="The name of the database")
    url: str = Field(
        default=...,
        description="The url string for the database connection. Can contain placeholders like ${MONGO_PASSWORD} that are then filled through .env parameters",
    )
    clear_on_initialize: bool = Field(
        default=False,
        description="If the entire database should be cleared when the instance is (re)started. This is useful for testing, and should be set to False in production.",
    )


class StorageSettings(BaseModel):
    """Configures the storage backend used by the instance."""

    provider: str = StorageProviderType.Mongodb.value
    database: DatabaseSettings = Field(default_factory=lambda: DatabaseSettings(url="sqlite:///instance.db"))
    page_size: int = Field(default=50, description="Number of items per page in paginated collections.")


class SecuritySettings(BaseModel):
    """Configures the security settings of the instance, including key rotation and http signatures."""

    mock_keycloak: bool = Field(default=False, description="If the keycloak connection should be mocked (for testing). Default: False.")
    key_rotation_enabled: bool = Field(
        default=False,
        description="If key rotation should be performed at all. Note that changing keys might ban an instance from certain Mastodon instances, so be careful with enabling key rotation.",
    )
    key_rotation_period: int = Field(default=30 * 24 * 3600, description="Number of seconds until actor keys are rotated.")
    key_rotation_overlap: int = 2
    key_size: int = Field(default=2048, description="Size of each actor key, in bytes.")
    signature_ttl: int = Field(default=300, description="Time-to-live, i.e. number of seconds a signature of a header is considered to be valid.")
    max_payload_size: int = 5_000_000  # 5MB
    allowed_algorithms: list[str] = Field(default_factory=lambda: ["rsa-sha256", "hs2019"])
    verify_ssl: bool = True  # TODO: implement ssl verification
    require_body_digest: bool = Field(
        default=True, description="If http signature verification also requires that the request body digest is included."
    )
    create_actor_from_token: bool = Field(
        default=True, description="If a local actor should be created automatically from the UserInfo if somebody logs into keycloak successfully"
    )


class FederationSettings(BaseModel):
    """Configures the instance and how it interacts with the federated network."""

    instance_name: str = "phederation"
    logging_prefix: str = Field(
        default="phederation", description="Will be added before every logger name. This can be used to distinguish admin and main instance logs."
    )
    auto_accept_follows_default: bool = Field(
        default=True,
        description="If follows should be accepted by default for actors that are created on this instance. The setting can be switched later, separately for each actor.",
    )
    delivery_timeout: int = Field(default=30, description="Timeout in seconds before a single delivery will fail, and then retry.")
    update_nodeinfo_period: int = Field(default=30 * 60, description="Number of seconds until the instance NodeInfo is updated.")
    delivery_queue_polling_period: int = Field(
        default=30, description="Number of seconds until new messages are polled from the delivery quue to send."
    )
    max_concurrent: int = Field(default=50, description="Maximum number of recipients to send to concurrently")
    retry_delay: int = Field(default=300, description="Delay in seconds until sending an activity is tried again (with exponential backoff).")
    max_retries: int = Field(default=3, description="Number of tries before sending an activity is considered failed.")
    cache_ttl: int = Field(default=60, description="Seconds for caching of actors and objects. No cache is created if this is set to zero.")
    cache_max: int = Field(default=10_000, description="Maximum number of objects in cache. No cache is created if this is set to zero.")
    delivery_queuing: str = Field(
        default=DeliveryQueueType.Automatic.value,
        description="Behavior for queueing of activity sending tasks. Immediate does not use a worker environment and send all messages directly, Worker will use temporalio workflows to queue delivery, and Automatic will choose between the two modes based on avaiability of the worker environment.",
    )
    max_concurrent_resolve: int = Field(default=50, description="Maximum number of ids to resolve concurrently in the resolve endpoint")
    endpoints: list[str] = Field(
        default_factory=lambda: ["sharedInbox", "objects", "media", "collections", "resolve"],
        examples=[["sharedInbox", "objects", "media"], ["media"]],
        description="Endpoints supported by the instance. With the current phederation software, this can be 'sharedInbox', 'objects', 'media', 'collections', 'resolve'.",
    )
    admin_mode: bool = Field(
        default=False,
        description="If True, puts the instance in admin mode. In this mode, no incoming requests are accepted if they do not originate from localhost/127.0.0.1, and additional API routes are added that allow maintenance for the instance.",
    )


class MediaSettings(BaseModel):
    """Configures how the Phederation instance handles media files.

    The settings here also allow to add higher security, including adding DataIntegrityProofs to objects and creating file hashes.
    """

    max_size: int = Field(default=10_000_000, description="Maximum file size for uploads that can be sent to the media endpoint.")
    create_file_hash: bool = Field(
        default=True,
        description="If set to True, will attach a SHA256-hash of file contents to the Document of media files, instead of signing just the 'content' field.",
    )
    proof_object_hash: bool = Field(
        default=False,
        description="If set to True, will attach a DataIntegrityProof object to all 'Document' objects, so that their integrity can be checked remotely. The proof only signs the 'hash' field of the 'Document', not the entire body.",
    )
    require_file_hash: bool = Field(
        default=False,
        description="If set to True, will verify that the SHA256-hash for files sent to the endpoint is correct. Processing fails with an exception if verification fails.",
    )
    require_content_hash: bool = Field(
        default=False,
        description="If set to True, will verify that the SHA256-hash of the object.content shared in client-to-server and server-to-server communication is attached and correct. Processing fails with an exception if verification fails.",
    )
    require_file_proof_in_inbox: bool = Field(
        default=False,
        description="If set to True, will verify the SHA256-hash of media shared in server2server communication using a DataIntegrityProof attached to the object. Processing fails with an exception if there is no DataIntegrityProof attached or if verification fails.",
    )
    allowed_types: list[str] = Field(
        default_factory=lambda: [
            "image/jpeg",
            "image/png",
            "image/gif",
            "video/mp4",
            "audio/mpeg",
            "text/txt",
            "text/json",
            "text/yaml",
            "application/json",
        ],
        description="Mime types of files that this endpoint accepts.",
    )


class StatisticsSettings(BaseModel):
    """Configures basic statistics logging in middleware. Disabled by default."""

    enabled: bool = True
    log_activities_incoming: bool = Field(
        default=True,
        description="If incoming activities from other instances should be logged. The content is not logged, just the type, actor_id, and the remote url",
    )
    log_activities_outgoing: bool = Field(
        default=True,
        description="If outgoing activities to other instances should be logged. The content is not logged, just the type, actor_id, and the remote url",
    )
    log_middleware: bool = Field(
        default=False,
        description="Log all middleware requests. This is very verbose and should only be used for debugging. It also is a security risk as user tokens can be logged, too",
    )
    disable_routes: list[str] = Field(default_factory=lambda: ["/auth", "/keys"])


class WorkerSettings(BaseModel):
    """Configures the interface to [temporalIO](https://temporal.io)."""

    enabled: bool = Field(default=True, description="If an asynchronous (temporalio) worker environment is available and should be used.")
    start_own_worker: bool = Field(
        default=True,
        description="If true, a single temporalio worker is started together with this instance. This is useful if no separate worker processes are started for scaling up, and the instance should handle all requests by itself.",
    )
    encrypt_in_transit: bool = Field(
        default=True,
        description="If enabled, encrypts the messages sent to and from the temporalio backend with a Fernet key. This means only the local instance can read the input and output, not a temporal server.",
    )
    tasks_queue_name: str = Field(default="temporal-workflows-queue", description="Name of the temporal worker task queue")


class PhedSettings(BaseSettings, case_sensitive=False):
    """Main settings class.

    Stores all types of settings, and is passed to all main classes (server, resolver, delivery, etc.) for configuration.
    """

    def __init__(self, yaml_file: str, env_file: str = ".env"):
        self.model_config["yaml_file"] = yaml_file
        self.model_config["env_file"] = env_file
        super().__init__()

    version: str = "0.3.3"
    domain: DomainSettings = Field(default=DomainSettings())
    storage: StorageSettings = Field(default=StorageSettings())
    security: SecuritySettings = Field(default=SecuritySettings())
    federation: FederationSettings = Field(default=FederationSettings())
    rate_limiter: RatelimiterSettings = Field(default=RatelimiterSettings())
    media: MediaSettings = Field(default=MediaSettings())
    statistics: StatisticsSettings = Field(default=StatisticsSettings())
    worker: WorkerSettings = Field(default=WorkerSettings())

    model_config: ClassVar[SettingsConfigDict] = SettingsConfigDict(
        yaml_file="settings/conf.yaml",
        env_prefix="PHEDERATION_",
        env_file=".env",
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
        yaml_file_encoding="utf-8",
        extra="ignore",
    )

    @classmethod
    @override
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """Customizes the file sources to load the settings.

        Note that the ORDER of the values returned here determines the priority of values loaded from different files.
        The first item in the list is the highest priority.

        Current order (highest first)
          - YAML file
          - ENV file
          - DotENV settings

        Args:
            settings_cls (type[BaseSettings]): Base settings class, used by YamlConfig.
            init_settings (PydanticBaseSettingsSource): Base class, not used.
            env_settings (PydanticBaseSettingsSource): Load settings from env file.
            dotenv_settings (PydanticBaseSettingsSource): Load settings from dotenv module.
            file_secret_settings (PydanticBaseSettingsSource): Load files from secrets.

        Returns:
            tuple[PydanticBaseSettingsSource, ...]: A list of file settings, in specific order.

        """
        return (
            YamlConfigSettingsSource(settings_cls),
            env_settings,
            dotenv_settings,
            init_settings,
            file_secret_settings,
        )



"""
Undo activity handler implementation.
"""

from typing import override

from phederation.models import ValidCollection, dereference
from phederation.models.activities import APActivity, APUndo
from phederation.models.objects import dereference_or_raise
from phederation.utils import ObjectId
from phederation.utils.base import AccessType, ActivityType, collection_id_from_name
from phederation.utils.exceptions import HandlerError, ValidationError

from .base import ActivityHandler


class UndoHandler(ActivityHandler):
    """Handle Undo activities."""

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Undo activity."""
        if activity.type != "Undo":
            raise ValidationError("Invalid activity type")

        if not activity.object:
            raise ValidationError("Missing object")

        if not activity.actor:
            raise ValidationError("Missing actor")

        # Validate object exists
        obj_dict_or_id = activity.object
        obj = await self.resolver.resolve_activity(obj_dict_or_id)

        # Validate actor has permission
        obj_actor = dereference(obj, key="actor")
        activity_actor = dereference(activity, key="actor")
        if not obj_actor or obj_actor != activity_actor:
            self.logger.warning(
                f"Actor {obj_actor} is not actor {activity_actor}, so not authorized to undo activity {obj.type}. Activity: {activity.id}, object={obj.id}"
            )
            raise ValidationError("Not authorized to undo activity")

        # Validate activity can be undone
        if obj.type not in ["Like", "Announce", "Follow", "Block"]:
            raise ValidationError(f"Cannot undo activity type: {obj.type}")

        return activity

    async def process_undo(self, activity: APActivity) -> APActivity:
        # Get original activity
        activity_to_undo_id = dereference(data=activity, key="object")
        if activity_to_undo_id:
            activity_to_undo_obj = await self.storage.activity.read(id=activity_to_undo_id)

            if not activity_to_undo_obj:
                self.logger.error(f"UndoHandler: Object with id {activity.object} does not exist locally. Failing silently.")
                return activity
        else:
            self.logger.warning(f"UndoHandler: activity {activity.type} does not have object")
            raise HandlerError(f"Activity {activity.type} does not have object")

        if not activity_to_undo_obj:
            self.logger.error(f"UndoHandler: Object with id {activity.object} does not exist locally. Failing silently.")
            return activity
        if not activity_to_undo_obj.id:
            self.logger.error(f"UndoHandler: Object of activity {activity.type} does not have proper ID. Failing silently.")
            return activity

        # Process based on activity type
        activity_type = ActivityType(activity_to_undo_obj.type)

        object_id = dereference(activity_to_undo_obj, key="object")
        actor_id = dereference(activity_to_undo_obj, key="actor")
        if not object_id or not actor_id:
            self.logger.error(f"UndoHandler: object_id={object_id} and actor_id={actor_id}. Both should be non-None. Failing silently.")
            return activity

        # the access type of all collections here is always "PRIVATE", since the actor itself is undoing something they did.
        if activity_type == ActivityType.LIKE:
            await self._undo_like(object_id=object_id, actor_id=actor_id)
        elif activity_type == ActivityType.ANNOUNCE:
            await self._undo_announce(object_id=object_id, actor_id=actor_id)
        elif activity_type == ActivityType.FOLLOW:
            await self._undo_follow(object_id=object_id, actor_id=actor_id, activity_id=activity_to_undo_obj.id)
        elif activity_type == ActivityType.BLOCK:
            await self._undo_block(object_id=object_id, actor_id=actor_id)
        return activity

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Process Undo activity."""
        activity = await self.process_undo(activity=activity)

        # Inform the recipients of the original activity about the undo
        await self._deliver_undo(activity)

        return activity

    async def _undo_like(self, object_id: ObjectId, actor_id: ObjectId) -> None:
        """Undo a Like activity."""
        collection_likes = collection_id_from_name(id=object_id, name=ValidCollection.Likes.value)
        _ = await self.collections.remove_from_collection(collection_id=collection_likes, items=[actor_id], access=AccessType.PRIVATE)
        collection_liked = collection_id_from_name(id=object_id, name=ValidCollection.Liked.value)
        _ = await self.collections.remove_from_collection(collection_id=collection_liked, items=[object_id], access=AccessType.PRIVATE)

    async def _undo_announce(self, object_id: ObjectId, actor_id: ObjectId) -> None:
        """Undo an Announce activity."""
        collection_shares_actor = collection_id_from_name(id=actor_id, name=ValidCollection.Shares.value)
        _ = await self.collections.remove_from_collection(collection_id=collection_shares_actor, items=[object_id], access=AccessType.PRIVATE)
        collection_shares_object = collection_id_from_name(id=object_id, name=ValidCollection.Shares.value)
        _ = await self.collections.remove_from_collection(collection_id=collection_shares_object, items=[actor_id], access=AccessType.PRIVATE)

    async def _undo_follow(self, object_id: ObjectId, actor_id: ObjectId, activity_id: ObjectId) -> None:
        """Undo a Follow activity."""
        # Remove from following collection
        if await self.storage.is_following(follower=actor_id, following=object_id):
            _ = await self.storage.follow.delete(id=activity_id)
            collection_following = collection_id_from_name(id=actor_id, name=ValidCollection.Following.value)
            _ = await self.collections.remove_from_collection(collection_id=collection_following, items=[activity_id], access=AccessType.PRIVATE)
            collection_followers = collection_id_from_name(id=object_id, name=ValidCollection.Followers.value)
            _ = await self.collections.remove_from_collection(collection_id=collection_followers, items=[activity_id], access=AccessType.PRIVATE)

        # TODO: make sure we actually remove the correct activity's follow.

    async def _undo_block(self, object_id: ObjectId, actor_id: ObjectId) -> None:
        """Undo a Block activity."""
        collection_blocks = collection_id_from_name(id=actor_id, name=ValidCollection.Blocks.value)
        _ = await self.collections.remove_from_collection(collection_id=collection_blocks, items=[object_id], access=AccessType.PRIVATE)

    async def _deliver_undo(self, activity: APActivity) -> None:
        """Deliver undo to recipients."""
        activity_to_undo_id = dereference_or_raise(activity, "object")
        # load the object from storage, instead of taking the one stored in the activity (for safety)
        activity_to_undo = await self.storage.activity.read(id=activity_to_undo_id)

        if not activity_to_undo:
            raise HandlerError(f"UndoHandler: Could not deserialize the activity to undo. Activity could not be found in storage: {activity_to_undo_id}")

        # Deliver to original recipients
        if activity_to_undo and activity_to_undo.to and len(activity_to_undo.to) > 0:
            result = await self.delivery.deliver_activity(
                activity=APUndo(id=activity.id, object=activity_to_undo.id, actor=activity.actor), recipients=activity_to_undo.to
            )
            self.logger.info(f"UndoHandler: Delivering object {activity_to_undo.id} to {activity.actor}. Status: {result.status_code}")

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        """Process Undo activity."""
        activity_processed = await self.process_undo(activity=activity)

        # Deliver undo to recipients (not done in inbox, only outbox)
        # self._deliver_undo(activity)

        return activity_processed.id



"""
actors.py
This module defines the Actor types for ActivityPub.
These actors are based on the ActivityStreams vocabulary and ActivityPub protocol.

For more information on Actors in ActivityPub, see:
https://www.w3.org/TR/activitypub/#actors
"""

from enum import Enum
from typing import Any, ClassVar, TypedDict, override

from pydantic import Field, HttpUrl, field_validator
from pydantic.config import ConfigDict
from pydantic_core import Url

from phederation.utils.base import ObjectId, APDateTime
from phederation.utils.exceptions import InvalidURLError

from .errors import BaseError
from .keys import APPublicKey
from .links import RelativeLink, RelativeLinkTemplate
from .objects import APObject, dereference, dereference_or_raise


class ErrorMessageActorUnauthorized(BaseError):
    message: str = "Unauthorized."
    description: str = "The actor is not authorized to perform this action."


class ActivityDict(TypedDict):
    actor: str
    object: dict[str, Any]
    published: str
    type: str
    context: str


class ActorType(Enum):
    """Actor types, subset of ObjectType."""

    ACTOR = "Actor"
    ACCOUNT = "Account"
    PROFILE = "Profile"
    USER = "User"

    PERSON = "Person"
    GROUP = "Group"
    ORGANIZATION = "Organization"
    SERVICE = "Service"
    APPLICATION = "Application"

    @override
    def __str__(self):
        return str(self.value)


class APAccount(APObject):
    type: str = ActorType.ACCOUNT.value
    username: str = Field(..., description="Username of the actor associated with this local account")
    fullname: str | None = Field(default=None, description="Actual name of the actor associated with this local account")
    email: str | None = Field(default=None, description="Email address, if available")
    created_at: APDateTime | None = Field(default=None, description="Created time")
    updated_at: APDateTime | None = Field(default=None, description="Last update time")

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "https://example.org/accounts/exampleuser",
                    "type": "Account",
                    "name": "exampleuser",
                    "fullname": "FirstnameExample LastnameExample",
                    "username": "exampleuser",
                    "email": "email@example.org",
                    "created_at": "2022-01-13 00:10:20.470276",
                }
            ]
        }
    )


class APActor(APObject):
    """
    Base Actor type as defined in ActivityPub.

    Specification: https://www.w3.org/TR/activitypub/#actor-objects
    Usage: https://www.w3.org/TR/activitypub/#actor-objects
    """

    type: str = ActorType.ACTOR.value
    inbox: ObjectId = Field(..., description="Primary inbox URL")
    outbox: ObjectId = Field(..., description="Primary outbox URL")
    shared_inbox: ObjectId | None = Field(default=None, description="Set if the actor instance has a shared inbox URL")
    following: None | ObjectId = Field(default=None, description="Following collection")
    followers: None | ObjectId = Field(default=None, description="Followers collection")
    liked: None | ObjectId = Field(default=None, description="Liked collection")
    blocks: None | ObjectId = Field(default=None, description="Collection for blocked objects/actors")
    objects: None | ObjectId = Field(default=None, description="Collection for objects created by the actor")
    collections: None | ObjectId = Field(default=None, description="Collection for collections created by the actor")
    streams: None | list[ObjectId] = Field(default=None, description="Additional streams")
    endpoints: None | dict[str, ObjectId] = Field(default=None, description="Additional endpoints")
    public_key: None | APPublicKey | str | list[str] | list[APPublicKey] = Field(default=None, description="Public key information, can be a list")
    local: bool = Field(default=True, description="If the actor is local to this instance.")
    discoverable: bool = Field(default=True, description="If this actor can be discovered in the /users route.")
    memorial: bool = Field(default=False, description="If this is a memorial of an actor.")

    # below are fields that can be updated. All fields above depend on the instance and the actor_id
    preferred_username: None | str = Field(default=None, description="Preferred username")
    autoaccept: bool = Field(default=True, description="If the actor automatically accepts follow requests.")

    @field_validator("inbox", "outbox", "following", "followers", "liked")
    @classmethod
    def validate_urls(cls, v: str | Url) -> ObjectId:
        try:
            return str(HttpUrl(v))
        except ValueError:
            raise InvalidURLError(f"Invalid URL: {v}")

    @override
    def serialize(self, include_context: bool = True) -> dict[str, Any]:
        """Serialize APActor object.
        This override of the super class is necessary to ensure the public key id is stored properly.

        Args:
            include_context (bool, optional): if the JSON-LD context should be included in the json dict. Defaults to True.

        Returns:
            dict[str, Any]: JSON-LD object for this actor.
        """
        actor_dict = super().serialize(include_context=include_context)
        key_ids = self.public_key_as_ids()
        if key_ids:
            actor_dict["key_id"] = key_ids
        return actor_dict

    def public_key_as_ids(self):
        public_key_ids = None
        if isinstance(self.public_key, str):
            public_key_ids = self.public_key
        if isinstance(self.public_key, APPublicKey):
            public_key_ids = dereference(self.public_key, "id")
        if isinstance(self.public_key, list):
            public_key_ids = [dereference_or_raise(key, "id") for key in self.public_key]
        return public_key_ids

    def add_public_key(self, key: ObjectId | APPublicKey):
        key_list: list[ObjectId] = []
        if isinstance(self.public_key, str) or isinstance(self.public_key, APPublicKey):
            key_list = [dereference_or_raise(self.public_key, "id"), dereference_or_raise(key, "id")]
        if isinstance(self.public_key, list):
            key_list = [dereference_or_raise(key, "id")] + [dereference_or_raise(key, "id") for key in self.public_key]
        self.public_key = key_list

    model_config: ClassVar[ConfigDict] = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "id": "https://example.org/users/exampleuser",
                    "type": "Person",
                    "name": "exampleuser",
                    "summary": "A test account for ActivityPub.",
                    "inbox": "http://example.org/users/exampleuser/inbox",
                    "outbox": "http://example.org/users/exampleuser/outbox",
                    "following": "http://example.org/users/exampleuser/following",
                    "followers": "http://example.org/users/exampleuser/followers",
                    "preferredUsername": "exampleuser",
                    "publicKey": {
                        "id": "http://example.org/keys/af7af0c4-fd87-4f8d-8f40-be130f464e4b",
                        "type": "Key",
                        "publicKeyPem": "-----BEGIN PUBLIC KEY-----\nDFGSDFJGSFJSDJFASDJASDQ121231+1231231SDFSFDGDFGD3453QARASD\n-----END PUBLIC KEY-----\n",
                    },
                    "links": [
                        {
                            "rel": "self",
                            "type": "application/activity+json",
                            "href": "http://example.org/users/exampleuser",
                        },
                        {
                            "rel": "http://ostatus.org/schema/1.0/subscribe",
                            "template": "http://example.org/authorize_interaction?uri={uri}",
                        },
                    ],
                    "discoverable": True,
                    "memorial": False,
                    "username": "exampleuser",
                    "domain": "example.org",
                    "autoaccept": "True",
                    "shared_inbox": "http://example.org/inbox",
                    "local": "True",
                    "created_at": "2022-01-13 00:10:20.470276",
                    "updated_at": "None",
                    "last_fetched_at": "None",
                }
            ]
        }
    )


class APPerson(APActor):
    """
    Represents a Person actor in ActivityPub.

    A Person is an individual user account.

    Properties:
        type (Literal["Person"]): Always set to "Person".

    Inherits all other properties from APActor.

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#dfn-person
    """

    type: str = Field(default=ActorType.PERSON.value, description="Indicates that this object represents a person")


class APGroup(APActor):
    """
    Represents a Group actor in ActivityPub.

    A Group represents a formal or informal collective of Actors.

    Attributes:
        type (str): Always set to "Group".

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#dfn-group
    """

    type: str = Field(default=ActorType.GROUP.value, description="Indicates that this object represents a group")


class APOrganization(APActor):
    """
    Represents an Organization actor in ActivityPub.

    An Organization is a kind of Actor representing a formal or informal organization.

    Attributes:
        type (str): Always set to "Organization".

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#dfn-organization
    """

    type: str = Field(
        default=ActorType.ORGANIZATION.value,
        description="Indicates that this object represents an organization",
    )


class APApplication(APActor):
    """
    Represents an Application actor in ActivityPub.

    An Application represents a software application.

    Attributes:
        type (str): Always set to "Application".

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#dfn-application
    """

    type: str = Field(
        default=ActorType.APPLICATION.value,
        description="Indicates that this object represents an application",
    )


class APService(APActor):
    """
    Represents a Service actor in ActivityPub.

    A Service is a kind of Actor that represents a service of any kind.

    Attributes:
        type (str): Always set to "Service".

    Specification: https://www.w3.org/TR/activitystreams-vocabulary/#dfn-service
    """

    type: str = Field(default=ActorType.SERVICE.value, description="Indicates that this object represents a service")


class APUser(APActor):
    """Schema for a user on this instance."""

    type: str = Field(default=ActorType.USER.value, description="Indicates that this object represents a user")
    links: list[RelativeLink | RelativeLinkTemplate]



"""
Media attachment handling.
"""

from collections.abc import Callable, Coroutine
from logging import Logger
from typing import Any

from fastapi.datastructures import UploadFile

from phederation.federation.resolver import ActivityPubResolver
from phederation.models import APActivity, APDocument, APObject
from phederation.models.objects import dereference
from phederation.models.proofs import DataIntegrityProof
from phederation.storage.base import StorageBackend
from phederation.utils import ActivityType, NodeInfo, ObjectId
from phederation.utils.exceptions import (
    AuthenticationError,
    HandlerError,
    MediaError,
    catch_exceptions,
)
from phederation.utils.logging import configure_logger
from phederation.utils.settings import PhedSettings


class MediaHandler:
    """Handle media attachments."""

    def __init__(self, storage_backend: StorageBackend, resolver: ActivityPubResolver, settings: PhedSettings):
        self.storage: StorageBackend = storage_backend
        self.resolver: ActivityPubResolver = resolver
        self.settings: PhedSettings = settings
        self.logger: Logger = configure_logger(__name__, prefix=self.settings.federation.logging_prefix)

    async def validate(self, activity: APObject, actor_id: ObjectId) -> APActivity:
        if not isinstance(activity, APActivity):
            raise HandlerError("activity send to media endpoint is not an APActivity")

        if activity.type != ActivityType.CREATE.value:
            raise HandlerError("Cannot create media without Create activity")

        if dereference(activity, "actor") != actor_id:
            raise AuthenticationError("Sending actor must be 'actor' in activity")

        document = activity.object
        if document is None:
            raise HandlerError("When sending media files, activity Create must have an object")
        if isinstance(document, APObject):
            # this is necessary to convert to a proper APDocument
            document = document.serialize()
        if isinstance(document, str):
            raise HandlerError("When sending media files, activity Create object must not be an url, but the actual object (as dict)")

        document = self.resolver.resolve_from_dict(document)
        if isinstance(document, DataIntegrityProof):
            raise HandlerError("Document to resolve was a DataIntegrityProof, not an APObject")
        if isinstance(document, NodeInfo):
            raise HandlerError("Document to resolve was a NodeInfo, not an APObject")
        document.attributed_to = actor_id
        activity.object = document

        if self.settings.media.require_content_hash:
            if not document.hash:
                raise HandlerError(
                    f"Document {(document.serialize())} should have a SHA256 hash of the file content in the 'hash' field.",
                    user_facing_message="Security settings demand that the SHA256 hash of the uploaded file is stored in the 'hash' field of the Document",
                )
        return activity

    @catch_exceptions(MediaError, "Failed to process attachment")
    async def process_attachment(
        self,
        attachment: UploadFile,
        description: APObject | dict[str, Any] | str | None,
        proof_string: Callable[..., Coroutine[Any, Any, DataIntegrityProof]] | None = None,
    ) -> APDocument:
        """
        Process media attachment.
        If required by security settings, also checks the file SHA-hash against the one provided in the description.

        Returns:
            Processed attachment object
        """
        if not description:
            description = {}
        if isinstance(description, str):
            description = {"description": description}
        media_info: dict[str, Any] = description.serialize() if isinstance(description, APObject) else description
        media_info["type"] = "Document"
        document = APDocument.deserialize(media_info)
        if not isinstance(document, APDocument):
            raise MediaError(f"Could not deserialize APDocument with description {description}")
        
        # store the document to get an id
        document.id = await self.storage.object.create(data=document, raise_if_exists=True)

        # Validate mime type
        mime_type = document.media_type or attachment.content_type
        if mime_type is None or mime_type not in self.settings.media.allowed_types:
            raise MediaError(f"Unsupported media type: {mime_type}")

        if attachment.size is None:
            raise MediaError("Media does not have set size")
        if attachment.size > self.settings.media.max_size:
            raise MediaError("Media too large")

        # Save file using storage backend
        file_id = await self.storage.save(document_id=document.id, content=attachment)

        # get metadata that was saved with the file
        file_metadata = await self.storage.metadata(file_id=file_id)

        # TODO: Process media based on type

        document.media_type = mime_type
        document.url = [file_id]
        document.name = file_metadata.filename

        if self.settings.media.require_file_hash:
            if document.hash != file_metadata.hash_sha256:
                raise MediaError(
                    f"SHA256 file hashes do not agree for document '{document.id}' with file '{file_id}'",
                    user_facing_message="SHA256 file hashes do not agree. Security settings demand that they are verified",
                )
        document.hash = file_metadata.hash_sha256

        if file_metadata.hash_sha256 is not None and proof_string is not None and self.settings.media.create_file_hash:
            document.proof = await proof_string(file_metadata.hash_sha256)
            document.proof.description = "Actor signed sha256 hash of the file content."

        # udpate the document with the new file information
        _ = await self.storage.object.upsert(id=document.id, data=document)

        return document



"""Like activity handler."""

from typing import override

from phederation.models import APObject, dereference
from phederation.models.activities import APActivity
from phederation.utils import ActivityType, ObjectId
from phederation.utils.base import AccessType, collection_id_from_name
from phederation.utils.exceptions import HandlerError, ValidationError

from .base import ActivityHandler


class LikeHandler(ActivityHandler):
    """Handles Like activities."""

    @override
    async def validate(self, activity: APActivity) -> APActivity:
        """Validate Like activity."""
        if not ActivityType.__contains__(activity.type) or ActivityType(activity.type) != ActivityType.LIKE:
            raise ValidationError("Invalid activity type")

        if not activity.actor:
            raise ValidationError("Like must have an actor")

        # TODO: check if actor is local to this instance

        if not activity.object:
            raise ValidationError("Like must have an object")

        return activity

    async def has_liked(self, actor_id: ObjectId, object_id: ObjectId):
        collection_liked = collection_id_from_name(id=actor_id, name="liked")
        return await self.collections.contains(collection_id=collection_liked, item=object_id)

    @override
    async def process_outbox(self, activity: APActivity) -> APActivity:
        """Handle Like activity."""
        actor_id = dereference(data=activity, key="actor")
        object_id = dereference(data=activity, key="object")

        if not actor_id or not object_id:
            raise HandlerError("Could not dereference actor or object id from activity")

        # Resolve target object
        obj = await self.resolver.resolve_object(object_id)

        # Check for duplicate like
        if await self.has_liked(actor_id, object_id):
            raise HandlerError("Already liked this object")

        # the access level is the maximum (more private) between activity and object visibility
        access1 = AccessType.from_visibility(activity.visibility)
        access2 = AccessType.from_visibility(obj.visibility)
        access = AccessType.maximum_security(access1, access2)

        # Store like
        collection_liked = collection_id_from_name(id=actor_id, name="liked")
        await self.collections.add_to_collection(collection_id=collection_liked, items=object_id, access=access)
        collection_likes = collection_id_from_name(id=object_id, name="likes")
        await self.collections.add_to_collection(collection_id=collection_likes, items=actor_id, access=access)

        # Notify object creator if it is not a local object, otherwise increase likes
        if getattr(obj, "local", False) == False:
            await self._notify_object_owner(obj, activity)

        self.logger.info(f"Handled Like activity: {actor_id} -> {object_id}")
        return activity

    async def _notify_object_owner(self, obj: APObject, activity: APActivity) -> None:
        """Notify object owner about the like."""
        if obj.attributed_to:
            target_actor = await self.resolver.resolve_actor(obj.attributed_to)
            if target_actor and target_actor.id and dereference(data=target_actor, key="inbox"):
                _ = await self.delivery.deliver_to_actor_inbox(activity=activity, actor_id=target_actor.id)
            else:
                self.logger.warning(f"LikeHandler: _notify_object_owner could not deliver to target actor: target_actor={target_actor.serialize()}")
        else:
            raise HandlerError("Object does not have an owner, cannot notify")

    @override
    async def process_inbox(self, activity: APActivity) -> None | str:
        # Actor and object are validated to be part of the activity.
        actor_id = dereference(data=activity, key="actor")
        object_id = dereference(data=activity, key="object")

        if not actor_id or not object_id:
            raise HandlerError("Could not dereference actor or object id from activity")

        #     Check that the object of the activity is an object on the local server.
        obj = await self.storage.object.read(id=object_id)
        if not obj:
            #  Otherwise, the Like activity can be safely treated like any other activity with no side effects.
            # https://www.w3.org/wiki/ActivityPub/Primer/Like_activity

            # Regardless, we will throw an exception here to inform the stream the object does not exist here.
            raise HandlerError(f"Object {object_id} does not exist on this instance.")

        #     Check that the actor of the activity is not the actor of an existing Like activity in the likes collection of the object.
        if await self.has_liked(actor_id=actor_id, object_id=object_id):
            return object_id

        #     Check that the responsible actor for the object has not blocked the actor of the Like.
        owner_id = dereference(obj, "attributedTo")
        if owner_id:
            collection_blocks = collection_id_from_name(id=owner_id, name="blocks")
            if await self.collections.contains(collection_id=collection_blocks, item=actor_id):
                return None

        #     Check that the actor has permission to view the object, such as being one of the addressees of the object or its Create activity.
        # TODO: implement visibility layers
        if obj.visibility != "public":
            raise HandlerError(f"Object {object_id} is not public.")

        # If these checks work:
        self.logger.warning(f"Storing like in likes collection of {object_id} in inbox")
        collection_liked = collection_id_from_name(id=actor_id, name="liked")
        await self.collections.add_to_collection(collection_id=collection_liked, items=object_id)
        collection_likes = collection_id_from_name(id=object_id, name="likes")
        await self.collections.add_to_collection(collection_id=collection_likes, items=actor_id)
